from .tool_colormap import tool_colormap
from .tool_plot_manipulation_functions import tool_zoom_plot_matplotlib
from .tool_math_functions import tool_multinom

from .tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib
from .tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly

from .tool_complete_field import tool_complete_field

from .tool_add_spacing_2D import tool_add_spacing_2D
from .tool_add_spacing_3D import tool_add_spacing_3D

from .tool_print_in_color import tool_print_in_color
from .tool_configure_axis import tool_configure_axis

from .tool_make_animation import tool_make_animation_movie, tool_make_animation_gif

from .tool_plotly_colorbar import tool_plotly_colorbar

from .tool_plotly_find_next_xN import tool_plotly_find_next_xN
from .tool_plotly_define_2D_plot_ax import tool_plotly_define_2D_plot_ax

from .tool_plotly_find_next_sceneN import tool_plotly_find_next_sceneN
from .tool_plotly_define_3D_plot_ax import tool_plotly_define_3D_plot_ax
from .tool_extract_node_arrays import tool_extract_node_arrays

from .tool_matplotlib_define_2D_plot_ax import tool_matplotlib_define_2D_plot_ax
from .tool_matplotlib_define_3D_plot_ax import tool_matplotlib_define_3D_plot_ax


from comfit.core.base_system_init import BaseSystemInit
from comfit.core.base_system_conf import BaseSystemConf
from comfit.core.base_system_evolve import BaseSystemEvolve
from comfit.core.base_system_calc import BaseSystemCalc
from comfit.core.base_system_plot import BaseSystemPlot
from comfit.core.base_system_get import BaseSystemGet

import scipy as sp

class BaseSystem(BaseSystemInit, BaseSystemConf, BaseSystemEvolve, BaseSystemCalc, BaseSystemPlot, BaseSystemGet):
    """
    The BaseSystem class is the base class for all systems in ComFiT
    """
    
    def fft(self, field):
        """
        Perform a fast Fourier transform on a field.

        Parameters
        ----------
        field : np.ndarray
            The field to transform
            
        Returns
        -------
        np.ndarray
            The transformed field

        Note
        ----
        The field is assumed to represent a field in the dimensions of the system.
        Thus, if the system is 2D and the field is 3D, it is assumed that 
        the provided field is a collection of 2D fields which will be transformed
        individually.
        """
        
        return sp.fft.fftn(field, axes=range(-self.dim, 0))

    def ifft(self, field):
        """Perform an inverse fast Fourier transform on a field.

        Parameters
        ----------
        field : np.ndarray
            The field to transform
            
        Returns
        -------
        np.ndarray
            The transformed field

        Note
        ----
        The field is assumed to represent a field in the dimensions of the system.
        Thus, if the system is 2D and the field is 3D, it is assumed that 
        the provided field is a collection of 2D fields which will be transformed
        individually.
        """
        
        return sp.fft.ifftn(field, axes=range(-self.dim, 0))


# Typing imports
from typing import TYPE_CHECKING, Any, Tuple
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# General packages
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

# Comfit packages
from comfit.plot import plot_field_matplotlib, plot_complex_field_matplotlib
from comfit.tool import tool_complete_field

def plot_angle_field_matplotlib(
        self: 'BaseSystem',
        angle_field: np.ndarray,
        **kwargs: Any
        ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot an angle field using matplotlib.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    angle_field : np.ndarray
        The angle field values in radians.
    \*\*kwargs : Any
        Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes objects containing the plot.
    """

    # Normalize around 0
    angle_field = np.mod(angle_field + np.pi, 2 * np.pi) - np.pi        

    if self.dim == 1:
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
        else:
            kwargs['vlim'] = [-np.pi, np.pi]
            kwargs['yticks'] = [-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]
            kwargs['yticklabels'] = [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$']

        
        return plot_field_matplotlib(self, angle_field, **kwargs)
    
    elif self.dim > 1:
        complex_field = np.exp(1j * angle_field)

        kwargs['plot_method'] = 'phase_angle'

        return plot_complex_field_matplotlib(self, complex_field, **kwargs)


from typing import Dict, List, Optional, Any, Tuple

import numpy as np
import plotly.graph_objects as go


def plot_subplots_plotly(
        number_of_rows: int,
        number_of_columns: int
        ) -> Tuple[go.Figure, Dict[str, Any]]:
    """Create a plotly figure with subplots.

    Parameters
    ----------
    number_of_rows : int
        Number of rows in the subplot grid.
    number_of_columns : int
        Number of columns in the subplot grid.

    Returns
    -------
    Tuple[go.Figure, Dict[str, Any]]
        A tuple containing the Plotly figure and axes dictionary.
    """
    fig = go.Figure()
    if number_of_columns == 1:
        axs = np.array([{'row': i+1, 'nrows': number_of_rows, 'col': 1, 'ncols': number_of_columns, 'colorbar': False} for i in range(number_of_rows)])
    elif number_of_rows == 1:
        axs = np.array([{'row': 1, 'nrows': number_of_rows, 'col': i+1, 'ncols': number_of_columns, 'colorbar': False} for i in range(number_of_columns)])
    else:
        axs = np.array([[{'row': i+1, 'nrows': number_of_rows, 'col': j+1, 'ncols': number_of_columns, 'colorbar': False} for j in range(number_of_columns)] for i in range(number_of_rows)])
    return fig, axs


from typing import Tuple, List, Union

import numpy as np

def tool_create_orthonormal_triad(
        t: Union[np.ndarray, List[float]]
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Create an orthonormal triad given a vector t.
    
    Parameters
    ----------
    t : np.ndarray
        Input vector of shape (3,). Does not need to be normalized.
        
    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple (e1, e2, t) containing the orthonormal triad vectors.
        Each vector is a normalized numpy array of shape (3,).
    """
    t = np.array(t)

    # Normalize the input vector
    t = t / np.linalg.norm(t)

    # Define two orthogonal vectors
    ey = np.array([0, 1, 0])
    ez = np.array([0, 0, 1])

    # Check the dot product of ey and t
    # If it is less than 0.8, we can use ey to create the orthogonal vector e1
    # Otherwise, we use ez to create e1
    if abs(np.dot(ey, t)) < 0.8:
        e1 = np.cross(ey, t)
    else:
        e1 = np.cross(ez, t)

    # Normalize e1
    e1 /= np.linalg.norm(e1)

    # Create the second orthogonal vector e2 by taking the cross product of t and e1
    e2 = np.cross(t, e1)

    # Return the orthonormal triad
    return e1, e2, t



from .plot_field_plotly import plot_field_plotly
from .plot_field_matplotlib import plot_field_matplotlib

from .plot_complex_field_matplotlib import plot_complex_field_matplotlib
from .plot_complex_field_plotly import plot_complex_field_plotly

from .plot_angle_field_matplotlib import plot_angle_field_matplotlib
from .plot_angle_field_plotly import plot_angle_field_plotly

from .plot_vector_field_plotly import plot_vector_field_plotly
from .plot_vector_field_matplotlib import plot_vector_field_matplotlib

from .plot_field_in_plane_plotly import plot_field_in_plane_plotly
from .plot_field_in_plane_matplotlib import plot_field_in_plane_matplotlib

from .plot_complex_field_in_plane_plotly import plot_complex_field_in_plane_plotly
from .plot_complex_field_in_plane_matplotlib import plot_complex_field_in_plane_matplotlib

from .plot_vector_field_in_plane_both_plot_libs import plot_vector_field_in_plane_both_plot_libs

from .plot_nodes_matplotlib import plot_nodes_matplotlib
from .plot_nodes_plotly import plot_nodes_plotly

from .plot_subplots_matplotlib import plot_subplots_matplotlib
from .plot_subplots_plotly import plot_subplots_plotly

from .plot_surface_matplotlib import plot_surface_matplotlib


from typing import TYPE_CHECKING, Any, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import scipy as sp
from mpl_toolkits.mplot3d import Axes3D
from skimage.measure import marching_cubes

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_colormap,
    tool_set_plot_axis_properties_matplotlib,
    tool_matplotlib_define_2D_plot_ax,
    tool_matplotlib_define_3D_plot_ax
)
from .plot_surface_matplotlib import plot_surface_matplotlib

def plot_complex_field_matplotlib(
        self: 'BaseSystem',
        complex_field: np.ndarray,
        **kwargs: Any
        ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot a complex field using matplotlib.

    Creates a visualization of a complex field using different plotting methods
    depending on the dimensionality of the field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    complex_field : np.ndarray
        The complex field to plot
    \*\*kwargs : Any
        Keyword arguments for the plot, see https://comfitlib.com/Plotting/

    Returns
    -------
    Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes objects containing the plot

    Examples
    --------
    >>> system.plot_complex_field_matplotlib(field, colorbar=True)
    """

    kwargs['colormap'] = kwargs.get('colormap', 'angle') # Override the default colormap with 'angle'
    complex_field, fig, ax, kwargs = self.plot_prepare(complex_field, field_type = 'complex', **kwargs)

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

        # Padding for the colorbar.
        padding=0.05

        ax.plot(self.x/self.a0, rho, color='black')

        # Color in the graph based on the argument of the complex field
        blend_factor=0.3 # The degree to which the color is blended with white
        cmap = kwargs['colormap_object']

        ax.fill_between([self.xmin/self.a0,(self.xmin+self.dx/2)/self.a0], [rho[0],(rho[0]+rho[1])/2],
                        color=(1-blend_factor)*np.array(cmap((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]), 
                        alpha=1)

        for i in range(1,self.xRes-1):
            ax.fill_between([(self.x[i]-self.dx/2)/self.a0,self.x[i]/self.a0], [(rho[i]+rho[i-1])/2,rho[i]],
                            color=(1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]), 
                            alpha=1)
            ax.fill_between([self.x[i]/self.a0,(self.x[i]+self.dx/2)/self.a0], [rho[i],(rho[i]+rho[i+1])/2],
                color=(1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]),  
                alpha=1)

        ax.fill_between([(self.xmax-1.5*self.dx)/self.a0,(self.xmax-self.dx)/self.a0], [(rho[-1]+rho[-2])/2,rho[-1]],
                        color=(1-blend_factor)*np.array(cmap((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]),  
                        alpha=1)
            
    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    elif self.dim == 2:
        # Keyword arguments particular to the 2D case
        plot_method = kwargs.get('plot_method', 'phase_angle')

        # Create a meshgrid
        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        if plot_method == '3Dsurface':
        
            # Padding for the colorbar
            padding=0.2

            # Keyword arguments particular to the 3D surface plot
            grid = kwargs.get('grid', True)
            kwargs['axis_equal'] = False
            
            ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
            kwargs['plot_is_3D'] = True
            
            # Get the colors from a colormap (e.g., hsv, but you can choose any other)
            colors = kwargs['colormap_object']((theta + np.pi) / (2 * np.pi))  # Normalizing theta to [0, 1]

            surf = ax.plot_surface(X/self.a0, Y/self.a0, rho, facecolors=colors)

        elif plot_method == 'phase_angle':
            

            # Keyword arguments particular to the phase angle plot
            grid = kwargs.get('grid', False)

            rho_normalized = rho / np.max(rho)

            custom_colormap = kwargs['colormap_object']
            
            # Check if an axis object is provided
            ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

            # Padding for the colorbar
            padding=0.05
            
            mesh = ax.pcolormesh(X/self.a0, Y/self.a0, theta, shading='auto', cmap=custom_colormap, vmin=-np.pi, vmax=np.pi)
            mesh.set_alpha(rho_normalized)
 

    elif self.dim == 3:

        grid = kwargs.get('grid', True)
        axis_equal = kwargs.get('axis_equal',True)

        plot_method = kwargs.get('plot_method', 'phase_blob')
        
        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        rho_normalized = rho / np.max(rho)

        colormap = kwargs['colormap_object']

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True
    
        if plot_method == 'phase_angle':
            
          
            
            # Padding for the colorbar
            padding=0.2

            for angle in [-2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3]:
                field_to_plot = theta.copy()
                field_to_plot[theta < angle - 1] = float('nan')
                field_to_plot[theta > angle + 1] = float('nan')
                field_to_plot[rho_normalized < 0.01] = float('nan')

                if np.nanmin(field_to_plot) < angle < np.nanmax(field_to_plot):
                    #TODO: make alpha a keyword argument (Vidar 11.03.24)
                    plot_surface_matplotlib(self, field=field_to_plot, 
                                        value=angle, 
                                        color=colormap((angle + np.pi) / (2 * np.pi)), 
                                        alpha=0.5,
                                        ax=ax)

            theta = np.mod(theta, 2 * np.pi)

            field_to_plot = theta.copy()
            field_to_plot[theta < np.pi - 1] = float('nan')
            field_to_plot[theta > np.pi + 1] = float('nan')
            field_to_plot[rho_normalized < 0.01] = float('nan')

            if np.nanmin(field_to_plot) < np.pi < np.nanmax(field_to_plot):
                #TODO: make alpha a keyword argument (Vidar 11.03.24)
                plot_surface_matplotlib(self, field=field_to_plot, 
                                        value=np.pi, 
                                        color=colormap(0), 
                                        alpha=0.5,
                                        ax=ax)
        
        elif plot_method == 'phase_blob':
            # TODO: change the function so that it used plot_surface_matplotlib (Vidar 11.03.24)
            # Padding for the colorbar
            padding=0.2
            
            phase_blob_threshold = kwargs.get('phase_blob_threshold', 0.5)

            if np.nanmin(rho_normalized)<phase_blob_threshold<np.nanmax(rho_normalized):
                verts, faces, _, _ = marching_cubes(rho_normalized, phase_blob_threshold)

                # Calculate the centroids of each triangle
                centroids = np.mean(verts[faces], axis=1)

                # Assuming theta is defined on the same grid as rho
                x, y, z = np.mgrid[0:rho_normalized.shape[0], 0:rho_normalized.shape[1], 0:rho_normalized.shape[2]]

                # Flatten the grid for interpolation
                points = np.c_[x.ravel(), y.ravel(), z.ravel()]
                theta_values = theta.ravel()

                # Interpolate theta at the vertices positions
                theta_faces = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')

                # Normalize theta values for color mapping
                theta_faces_normalized = (theta_faces + np.pi) / (2*np.pi)

                # Map normalized theta values to colors
                colors = colormap(theta_faces_normalized)
                    
                # Plot the complex field
                ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                faces, 
                                (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                facecolor=colors, antialiased=False)
            
                # Plot the shadows on the edges
                plot_shadows = kwargs.get('plot_shadows', True)
                if plot_shadows:
                    ax.plot_trisurf((self.xmin+0*verts[:, 0]*self.dx)/self.a0, 
                                    (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                    faces, 
                                    (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                    facecolor='black', antialiased=True,
                                    alpha=0.1)

                    ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                    (self.ymax+0*verts[:, 1]*self.dy)/self.a0, 
                                    faces, 
                                    (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                    facecolor='black', antialiased=True,
                                    alpha=0.1)
                    
                    ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                    (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                    faces, 
                                    (self.zmin+0*verts[:, 2]*self.dz)/self.a0, 
                                    facecolor='black', antialiased=True,
                                    alpha=0.1)



    if kwargs['colorbar']:

        mappable = plt.cm.ScalarMappable(cmap=kwargs['colormap_object'])
        mappable.set_array([])
        mappable.set_clim(-np.pi, np.pi)
        cbar = plt.colorbar(mappable, ax=ax, pad=padding)

        cticks = kwargs.get('cticks', [-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi])
        cbar.set_ticks(cticks)

        cticklabelse = kwargs.get('cticklabels', [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])
        cbar.set_ticklabels(cticklabelse)

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)

    return fig, ax



from typing import Union, Tuple, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

def tool_complete_field(
        self : 'BaseSystem', 
        field : np.ndarray
        ) -> np.ndarray:
    """Extends a partial field array to a complete array by tiling.

    Parameters
    ----------
    self : 'BaseSystem'
        A BaseSystem (or derived) instance.
    field : np.ndarray
        Input field array to be extended. Can be 1D, 2D, or 3D array with partial dimensions.

    Returns
    -------
    np.ndarray
        Extended field array with complete dimensions matching self.xRes, self.yRes, and self.zRes.

    Notes
    -----
    If input field is incomplete (has dimensions of 1 in some axes), it will be tiled
    to match the full resolution specified by self.xRes, self.yRes, and self.zRes.
    """

    # 2 dimensional fields
    if field.shape == (self.xRes,1):
        field = np.tile(field,(1,self.yRes))

    elif field.shape == (1,self.yRes):
        field = np.tile(field,(self.xRes,1))

    # 3 dimensional fields
    elif field.shape == (self.xRes,1,1):
        field = np.tile(field,(1,self.yRes,1))
        field = np.tile(field,(1,1,self.zRes))

    elif field.shape == (1,self.yRes,1):
        field = np.tile(field,(self.xRes,1,1))
        field = np.tile(field,(1,1,self.zRes))

    elif field.shape == (1,1,self.zRes):
        field = np.tile(field,(self.xRes,1,1))
        field = np.tile(field,(1,self.yRes,1))

    elif field.shape == (self.xRes,self.yRes,1):
        field = np.tile(field,(1,1,self.zRes))

    elif field.shape == (self.xRes,1,self.zRes):
        field = np.tile(field,(1,self.yRes,1))

    elif field.shape == (1,self.yRes,self.zRes):
        field = np.tile(field,(self.xRes,1,1))

    return field


from typing import Tuple

#General packages
import numpy as np

def tool_add_spacing_2D(
        X,
        Y,
        U,
        V,
        spacing
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Add spacing to a 2D field.

    Parameters
    ----------
    X : np.ndarray
        The x-coordinates.
    Y : np.ndarray
        The y-coordinates.
    U : np.ndarray
        The x-components of the field.
    V : np.ndarray
        The y-components of the field.
    spacing : int
        The spacing to be added.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        The spaced x-coordinates, y-coordinates, x-components of the field, and y-components of the field.
    """

    X = X[::spacing, ::spacing]
    Y = Y[::spacing, ::spacing]
    U = U[::spacing, ::spacing]
    V = V[::spacing, ::spacing]
    return X,Y,U,V


from typing import Union, Optional, Literal, Tuple, Dict, List, Any, TYPE_CHECKING


import numpy as np
import scipy as sp

from comfit.tool import tool_colormap
from comfit.tool import tool_create_orthonormal_triad
from comfit.tool import tool_set_plot_axis_properties_matplotlib
from comfit.tool import tool_complete_field
from comfit.tool import tool_print_in_color
from comfit.tool import tool_matplotlib_define_3D_plot_ax
from comfit.tool import tool_plotly_define_3D_plot_ax

from comfit.plot import plot_surface_matplotlib
from comfit.plot import plot_field_matplotlib
from comfit.plot import plot_field_plotly
from comfit.plot import plot_complex_field_matplotlib
from comfit.plot import plot_complex_field_plotly
from comfit.plot import plot_angle_field_matplotlib
from comfit.plot import plot_angle_field_plotly
from comfit.plot import plot_vector_field_matplotlib
from comfit.plot import plot_vector_field_plotly
from comfit.plot import plot_field_in_plane_matplotlib
from comfit.plot import plot_field_in_plane_plotly
from comfit.plot import plot_complex_field_in_plane_matplotlib
from comfit.plot import plot_complex_field_in_plane_plotly
from comfit.plot import plot_nodes_matplotlib
from comfit.plot import plot_nodes_plotly
from comfit.plot import plot_vector_field_in_plane_both_plot_libs

from comfit.plot import plot_subplots_matplotlib
from comfit.plot import plot_subplots_plotly

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.tri as mtri
import matplotlib.cm as cm
import matplotlib.axes
import matplotlib.figure

import mpl_toolkits

import plotly.graph_objects as go
import plotly.figure_factory as ff
import plotly.colors as pc

from skimage.measure import marching_cubes

import time

class BaseSystemPlot:
    """Plotting methods for the base system class"""

    def plot_prepare(
        self: 'BaseSystemPlot', 
        field: np.ndarray, 
        field_type: Literal['real', 'complex', 'angle', 'vector'], 
        **kwargs: any
        ) -> Tuple[np.ndarray, Union[go.Figure, plt.Figure], Any, Dict]:
        """Prepare axis and figure for plotting.

        Parameters
        ----------
        field : np.ndarray
            Field to be plotted.
        field_type : str
            Type of the field to be plotted.

        Returns
        -------
        Tuple[np.ndarray, Union[go.Figure, plt.Figure], Any, Dict]
            Tuple containing the field, figure (plotly or matplotlib), 
            axis (plotly or matplotlib) and keyword arguments for plotting.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)
        
        ## Define figure
        if plot_lib == "plotly":
            fig = kwargs.get('fig', go.Figure())
            ax = kwargs.get('ax', {'row': 1, 'col': 1, 'nrows': 1, 'ncols': 1, 'colorbar': False})

        elif plot_lib == "matplotlib":
            fig = kwargs['fig'] if 'fig' in kwargs else plt.figure()
            ax = kwargs.get('ax', None)

        # Extend field if not a complete array is given
        if field_type in ['real', 'complex', 'angle']:
            field = tool_complete_field(self, field)
        elif field_type == 'vector':
            field_copy = []
            for n in range(len(field)):
                field_copy.append(tool_complete_field(self, field[n]))
            field = np.array(field_copy)

        ## Checks on field
        # Check if the provided field is bool
        if field.dtype == bool:
            field = field.astype(float)

        # Check if the field is nan
        kwargs['field_is_nan'] = False
        if np.all(np.isnan(field)):
            kwargs['field_is_nan'] = True

        # Check if the field is an order parameter containing several components
        if field_type in ['real', 'complex', 'angle']:
            if field.ndim == self.dim+1:
                tool_print_in_color('Warning: The provided field seems to be an order parameter containing several fields. Only the zeroth component will be plotted.')
                field = field[0]

        # Check if the field is complex when it should be real
        if field_type in ['real', 'vector', 'angle']:
            if np.iscomplexobj(field):
                tool_print_in_color("Warning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.")
                print('Max imaginary part: ', np.max(np.imag(field)))
                field = np.real(field)

        ## Define default kwarg values
        kwargs['colormap'] = kwargs.get('colormap', 'viridis')
        kwargs['colormap_object'] = tool_colormap(kwargs['colormap'], plot_lib=plot_lib)

        colorbar_default = False
        if field_type == 'complex':
            colorbar_default = True
        if field_type in ['real','angle'] and self.dim>1:
            colorbar_default = True
        if field_type == 'vector' and plot_lib == 'plotly':
            colorbar_default = True
        kwargs['colorbar'] = kwargs.get('colorbar',  colorbar_default)

        kwargs['vlim_symmetric'] = kwargs.get('vlim_symmetric', False)

        grid_default = False
        if field_type == 'real' and self.dim in [1,3]:
            grid_default = True
        kwargs['grid'] = kwargs.get('grid', grid_default)

        axis_equal_default = False
        if self.dim in [2,3]:
            axis_equal_default = True
        kwargs['axis_equal'] = kwargs.get('axis_equal', axis_equal_default)

        kwargs['plot_is_3D'] = True if self.dim == 3 else False

        if field_type == 'real':
            # Field limits
            field_min = np.min(field)
            field_max = np.max(field)

            if 'vlim' in kwargs:
                vlim = kwargs['vlim']
                vmin = vlim[0]
                vmax = vlim[1]
            else:
                vmin = field_min
                vmax = field_max
                if 'vlim_symmetric' in kwargs:
                    if kwargs['vlim_symmetric']:
                        vmax = max(abs(vmin), abs(vmax))
                        vmin = -vmax
            
            if plot_lib == 'plotly':
                ax['vmin'] = min(vmin, ax.get('vmin', vmin)) 
                ax['vmax'] = max(vmax, ax.get('vmax', vmax))

            else:
                kwargs['vmin'] = vmin
                kwargs['vmax'] = vmax

        return field, fig, ax, kwargs

    def plot_field(
            self: 'BaseSystemPlot', 
            field: np.ndarray, 
            **kwargs: Any
            ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot a field.

        Parameters
        ----------
        field : np.ndarray
            Field to be plotted.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ 
            for a full list of keyword arguments.

        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes 
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default 
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_field_plotly or plot_field_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """
        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_field_plotly(self, field, **kwargs)
        elif plot_lib == "matplotlib":
            return plot_field_matplotlib(self, field, **kwargs)

    def plot_complex_field(
            self: 'BaseSystemPlot', 
            complex_field: np.ndarray, 
            **kwargs: Any
            ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot a complex field.

        Parameters
        ----------
        complex_field : np.ndarray
            Complex field to be plotted.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).
        
        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_complex_field_plotly or plot_complex_field_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_complex_field_plotly(self, complex_field, **kwargs)
        elif plot_lib == "matplotlib":
            return plot_complex_field_matplotlib(self, complex_field, **kwargs)

    def plot_angle_field(
            self: 'BaseSystemPlot', 
            angle_field: np.ndarray, 
            **kwargs: Any
            ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot an angle field.

        Parameters
        ----------
        angle_field : np.ndarray
            Angle field to be plotted.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_angle_field_plotly or plot_angle_field_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_angle_field_plotly(self, angle_field, **kwargs)
        elif plot_lib == "matplotlib":
            return plot_angle_field_matplotlib(self, angle_field, **kwargs)

    def plot_vector_field(
            self: 'BaseSystemPlot', 
            vector_field: np.ndarray, 
            **kwargs: Any
            ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot a vector field.

        Parameters
        ----------
        vector_field : np.ndarray
            Vector field to be plotted.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_vector_field_plotly or plot_vector_field_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_vector_field_plotly(self, vector_field, **kwargs)
        elif plot_lib == "matplotlib":
            return plot_vector_field_matplotlib(self, vector_field, **kwargs)

    def plot_field_in_plane(
        self: 'BaseSystemPlot',
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot the field in a plane.

        Parameters
        ----------
        field : np.ndarray
            Field to be plotted.
        normal_vector : np.ndarray, optional
            Normal vector of the plane. If None, the normal vector will be calculated.
        position : np.ndarray, optional
            Position of the plane. If None, the position will be calculated.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_field_in_plane_plotly or plot_field_in_plane_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_field_in_plane_plotly(
                self, field, normal_vector, position, **kwargs
            )
        elif plot_lib == "matplotlib":
            return plot_field_in_plane_matplotlib(
                self, field, normal_vector, position, **kwargs
            )

    def plot_complex_field_in_plane(
        self: 'BaseSystemPlot',
        complex_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot the complex field in a plane.

        Parameters
        ----------
        complex_field : np.ndarray
            Complex field to be plotted.
        normal_vector : np.ndarray, optional
            Normal vector of the plane. If None, the normal vector will be calculated.
        position : np.ndarray, optional
            Position of the plane. If None, the position will be calculated.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.

        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_complex_field_in_plane_plotly or plot_complex_field_in_plane_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)


        if plot_lib == "plotly":
            return plot_complex_field_in_plane_plotly(
                self, complex_field, normal_vector, position, **kwargs
            )
        elif plot_lib == "matplotlib":
            return plot_complex_field_in_plane_matplotlib(
                self, complex_field, normal_vector, position, **kwargs
            )

    def plot_angle_field_in_plane(
        self: 'BaseSystemPlot',
        angle_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot the angle field in a plane.

        Parameters
        ----------
        angle_field : np.ndarray
            Angle field to be plotted.
        normal_vector : np.ndarray, optional
            Normal vector of the plane. If None, the normal vector will be calculated.
        position : np.ndarray, optional
            Position of the plane. If None, the position will be calculated.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_angle_field_plotly or plot_angle_field_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)


        complex_field = np.exp(1j * angle_field)
        return self.plot_complex_field_in_plane(complex_field, normal_vector=normal_vector, position=position, **kwargs)

    def plot_vector_field_in_plane(
        self: 'BaseSystemPlot',
        vector_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        spacing = None,
        **kwargs: Any
        ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot the vector field in a plane.

        Parameters
        ----------
        vector_field : np.ndarray
            Vector field to be plotted.
        normal_vector : np.ndarray, optional
            Normal vector of the plane. If None, the normal vector will be calculated.
        position : np.ndarray, optional
            Position of the plane. If None, the position will be calculated.
        spacing : float, optional
            Spacing between the vectors.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).

        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the plot_vector_field_both_plot_libs function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)


        return plot_vector_field_in_plane_both_plot_libs(self, vector_field, normal_vector, position, spacing, **kwargs)

    def plot_nodes(
            self: 'BaseSystemPlot', 
            nodes: Dict[str, Any], 
            **kwargs: Any
            ) ->  Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot nodes.

        Parameters
        ----------
        nodes : np.ndarray
            Nodes to be plotted.
        \*\*kwargs : Any
            Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).
        
        Note
        ----
        plot_lib is a possible keyword argument. If not provided, the default
        plotting library (plotly or matplotlib) will be used.
        See the corresponding plot_nodes_plotly or plot_nodes_matplotlib function
        in the Auxiliary Plot Functions documentation for further details.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == 'plotly':
            return plot_nodes_plotly(self, nodes, **kwargs)
        elif plot_lib == 'matplotlib':
            return plot_nodes_matplotlib(self, nodes, **kwargs)


    # Figure handling methods
    def plot_subplots(
            self: 'BaseSystemPlot', 
            number_of_rows: int, 
            number_of_columns: int, 
            **kwargs: Any
            ) -> Tuple[Union[go.Figure, plt.Figure], Any]:
        """Plot subplots.

        Parameters
        ----------
        number_of_rows : int
            Number of rows in the subplot.
        number_of_columns : int
            Number of columns in the subplot.
        
        Returns
        -------
        Tuple[Union[go.Figure, plt.Figure], Any]
            Tuple containing the figure (plotly or matplotlib) and the axes
            dictionary (plotly or matplotlib).
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == "plotly":
            return plot_subplots_plotly(number_of_rows, number_of_columns)
        elif plot_lib == "matplotlib":
            return plot_subplots_matplotlib(number_of_rows, number_of_columns, **kwargs)

    def plot_save(
        self: 'BaseSystemPlot', 
        fig: Union[plt.Figure, go.Figure], 
        counter: Optional[int] = None, 
        **kwargs: Any
        ) -> None:
        """Save a figure.

        Parameters
        ----------
        fig : Union[plt.Figure, go.Figure]
            Figure to be saved.
        counter : int, optional
            Counter for the figure. If None, the figure will be saved as 'plot.png'.
        \*\*kwargs : Any
            Optional arguments: ID, image_size_inches, dpi.
        
        Returns
        -------
        None
            Save the figure. 
        """

        # Check if the figure is the first argument (deprecated)
        if isinstance(fig, int):
            tool_print_in_color("Warning: The plot_save method has been called with the figure as the second argument. This is deprecated. Please call the method with the figure as the first argument, then counter.")
            counter_tmp = fig
            fig = counter
            counter = counter_tmp
        
        # Get ID
        ID=kwargs.get('ID', None)

        # Keyword arguments
        image_size_inches=kwargs.get('image_size_inches', (6,5))
        dpi=kwargs.get('dpi', 100)

        # Set filename
        if counter is None:
            if ID is None:
                filename = 'plot.png'
            else:
                filename = f'plot_{ID}.png'
        else:
            if ID is None:
                filename = f'plot_{counter}.png'
            else:
                filename = f'plot_{counter}_{ID}.png'


        # Save the figure
        # Determine the plotting library based on figure type
        plot_lib = "plotly" if isinstance(fig, go.Figure) else "matplotlib"

        if plot_lib == "plotly":
            fig.write_image(filename)

        elif plot_lib == "matplotlib":
            fig.set_size_inches(image_size_inches)
            fig.savefig(filename, dpi=dpi)
            plt.close(fig)


    def show(
            self: 'BaseSystemPlot', 
            fig: Union[plt.Figure, go.Figure]
            ) -> None:
        """Show a figure.

        Parameters
        ----------
        fig : Union[plt.Figure, go.Figure]
            Figure to be shown.
        
        Returns
        -------
        None
            Show the figure.
        """

        plot_lib = "plotly" if isinstance(fig, go.Figure) else "matplotlib"

        if plot_lib == "matplotlib":
            plt.show()
        elif plot_lib == "plotly":
            fig.show()


from typing import TYPE_CHECKING, Optional, Any, Tuple, Dict
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
from scipy.interpolate import griddata
from skimage.measure import marching_cubes

# Third-party library imports
import plotly.graph_objects as go

# Local imports
from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly
from comfit.tool import tool_plotly_define_3D_plot_ax, tool_plotly_colorbar

def plot_field_in_plane_plotly(
        self: 'BaseSystem',
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> Tuple[go.Figure, Dict]:
    """Plot the field in a plane perpendicular to the given normal vector.
    
    Uses scipy.interpolate.griddata and plotly's Mesh3d for visualization.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    field : np.ndarray
        The field to be plotted
    normal_vector : np.ndarray, optional
        The normal vector of the plane. Defaults to [0,1,0]
    position : np.ndarray, optional
        The position of the plane. Defaults to the middle of the system
    \*\*kwargs : Any
        Additional keyword arguments for plot customization. See https://comfitlib.com/Plotting/

    Returns
    -------
    Tuple[go.Figure, dict]
        The figure object and axes dictionary containing the plot

    Raises
    ------
    Exception
        If the field dimension is not 3D
    """

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    field, fig, ax, kwargs = self.plot_prepare(field, field_type = 'real', **kwargs)

    ax = tool_plotly_define_3D_plot_ax(fig, ax) #Defines sceneN, plot_dimension


    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector = [0, 1, 0]

    normal_vector = np.array(normal_vector) / np.linalg.norm(normal_vector)
    height_above_plane = (self.x - position[0]) * normal_vector[0] + (self.y - position[1]) * normal_vector[1] + (self.z - position[2]) * normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    field_values = field.ravel()

    # Interpolate field at the vertices positions
    field_verts = griddata(points, field_values, centroids, method='nearest')

    ax['vmin'] = kwargs.get('vmin', np.min(field_verts))
    ax['vmax'] = kwargs.get('vmax', np.max(field_verts))

    # Add trace
    fig.add_trace(go.Mesh3d(
        x=(self.xmin + verts[:, 0] * self.dx) / self.a0,
        y=(self.ymin + verts[:, 1] * self.dy) / self.a0,
        z=(self.zmin + verts[:, 2] * self.dz) / self.a0,
        i=faces[:, 0],
        j=faces[:, 1],
        k=faces[:, 2],
        intensity=field_verts,  
        intensitymode='cell',  
        colorscale=kwargs['colormap_object'],
        showscale=False,
        scene=ax['sceneN']
    ))

    if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='normal'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)
    return fig, ax



from typing import List, Optional, Tuple, Union

# Local imports
from .tool_print_in_color import tool_print_in_color

def tool_configure_axis(
        dim: int, 
        name: str, 
        xlim: Optional[List[float]]=None, 
        xmin: Optional[float]=None, 
        xmax: Optional[float]=None, 
        xRes: Optional[int]=None, 
        dx: Optional[float]=None
        ) -> Tuple[List[float], float, float, int, float]:
    """Configure a single axis (x, y, or z) based on the provided parameters.

    When constructing a grid, providing all parameters will over-specify the axis.
    This function takes care of the hierarchy of values and determines the axis limits,
    minimum value, maximum value, resolution, and step size.

    The hierarchy of values is as follows:
    xlim > xmin > xmax > xRes > dx

    Parameters
    ----------
    dim : int
        Dimension of the system (1, 2, or 3)
    name : str
        Name of the axis ('x', 'y', 'z') for error messages
    xlim : list of float, optional
        A 2-element list [min, max] for the axis limits
    xmin : float, optional
        Minimum value of the axis
    xmax : float, optional 
        Maximum value of the axis
    xRes : int, optional
        Resolution (number of points) along the axis
    dx : float, optional
        Step size (spacing between points) along the axis

    Returns
    -------
    tuple
        (xlim, xmin, xmax, xRes, dx) where:
        - xlim is a list of [min, max] values
        - xmin is the minimum value (float)
        - xmax is the maximum value (float) 
        - xRes is the resolution (int)
        - dx is the step size (float)

    Notes
    -----
    Even though the function takes parameters 'x'-specific parameters, 
    it is used for 'y' and 'z' as well.
    """

    # Find the axis limits
    xlim_YUP_provided = xlim is not None
    xlim_NOT_provided = xlim is None
    
    xmin_YUP_provided = xmin is not None
    xmin_NOT_provided = xmin is None

    xmax_YUP_provided = xmax is not None
    xmax_NOT_provided = xmax is None

    xRes_YUP_provided = xRes is not None
    xRes_NOT_provided = xRes is None

    dx_YUP_provided = dx is not None
    dx_NOT_provided = dx is None

    if xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. no values provided

        xmin = 0
        xmax = 101 
        xRes = 101
        dx = 1.0

        if (name == 'y' and dim < 2) or (name == 'z' and dim < 3):
            xmax = 1
            xRes = 1
            
        xlim = [xmin, xmax]
        
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. dx = 0.1

        xmin = 0
        xmax = 101
        xRes = round((xmax - xmin) / dx)
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xRes = 56

        xmin = 0
        xmax = xRes
        dx = (xmax - xmin) / xRes
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xRes = 56, dx = 0.1
        
        xmin = 0
        xmax = xmin + xRes * dx
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xmax = 56
        
        xmin = 0 if xmax > 0 else xmax-101
        xRes = 101
        dx = (xmax - xmin) / xRes
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xmax = 21, dx = 0.1

        xmin = 0 if xmax > 0 else xmax-101
        xRes = round((xmax - xmin) / dx)
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xmax = 21, xRes = 56

        xmin = 0 if xmax > 0 else xmax-xRes
        dx = (xmax - xmin) / xRes
        xlim = [xmin, xmax]

    elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xmax = 21, xRes = 56, dx = 0.1

        xmin = xmax - xRes * dx
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xmin = -5

        xmax = xmin+101
        xRes = 101
        dx = 1.0
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xmin = -5, dx = 0.1

        xRes = 101
        xmax = xmin + xRes * dx
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xmin = -5, xRes = 56

        dx = 1.0
        xmax = xmin + xRes * dx
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xmin = -5, xRes = 56, dx = 0.1

        xmax = xmin + xRes * dx
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xmin = -5, xmax = 21

        xRes = xmax - xmin
        dx = 1.0
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xmin = -5, xmax = 21, dx = 0.1

        xRes = round((xmax - xmin) / dx)
        dx = (xmax - xmin) / xRes 
        xlim = [xmin, xmax]

    elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xmin = -5, xmax = 21, xRes = 56

        dx = (xmax - xmin) / xRes
        xlim = [xmin, xmax]
    
    elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xmin = -5, xmax = 21, xRes = 56, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}min, {name}max, and {name}Res trumps providing d{name}.', 'yellow')
        dx = (xmax - xmin) / xRes
        xlim = [xmin, xmax]
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21]

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = 101
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], dx = 0.1

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = round((xmax - xmin) / dx)
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xRes = 56

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xRes = 56, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim and xRes trumps providing d{name}.', 'yellow')
        
        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmax = 21

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = 101
        dx = (xmax - xmin) / xRes

    elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmax = 21, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = round((xmax - xmin) / dx)
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmax = 21, xRes = 56

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmax = 21, xRes = 56, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')
        tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes

    elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmin = -5

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = 101
        dx = (xmax - xmin) / xRes

    elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmin = -5, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = round((xmax - xmin) / dx)
        dx = (xmax - xmin) / xRes

    elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xRes = 56

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xRes = 56, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')
        tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xmax = 21

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = 101
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        xRes = round((xmax - xmin) / dx)
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, xRes = 56

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
        # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, xRes = 56, dx = 0.1

        tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')
        tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

        xmin = xlim[0]
        xmax = xlim[1]
        dx = (xmax - xmin) / xRes
    
    return xlim, xmin, xmax, xRes, dx


from typing import Optional, Union

# General packages
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt

def tool_matplotlib_define_3D_plot_ax(
        fig : plt,
        ax : Optional[Union[Axes3D, plt.Axes]] = None
        ) -> Axes3D:
    """
    Defines the axes for a 3D plot in matplotlib.
    
    Parameters
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    ax : matplotlib.axes._subplots.Axes3DSubplot
        Axes object.
    
    Returns
    -------
    ax : matplotlib.axes._subplots.Axes3DSubplot
        Axes object.
    """

    if ax is not None and ax is not isinstance(ax, Axes3D):
        # Get row and column
        subplotspec = ax.get_subplotspec()
        gridspec = subplotspec.get_gridspec()
        row = subplotspec.rowspan.start
        col = subplotspec.colspan.start

        ax.remove()

        fig.add_subplot(gridspec[row, col], projection='3d')
        ax = fig.get_axes()[-1]

    if ax is None:
        ax = fig.add_subplot(111, projection='3d')
    
    return ax


from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import numpy as np
from skimage.measure import marching_cubes

def plot_surface_matplotlib(
        self: 'BaseSystem',
        **kwargs: Any
        ) -> Axes:
    """Plot the surface of the given field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    \*\*kwargs : Any
        field : ndarray
            3D array containing the field values
        value : float
            Isosurface value
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. Defaults to current axes.
        alpha : float, optional
            Transparency value. Defaults to 0.5.
        color : str, optional
            Surface color. Defaults to 'b'.

    Returns
    -------
    matplotlib.axes.Axes
        The axes containing the surface plot.
    """
    field = kwargs['field']
    value = kwargs['value']
    ax = kwargs.get('ax', plt.gca())
    
    alpha = kwargs.get('alpha', 0.5)
    color = kwargs.get('color', 'b')

    verts, faces, _, _ = marching_cubes(field, value)
    x = (self.xmin+verts[:, 0]*self.dx)/self.a0
    y = (self.ymin+verts[:, 1]*self.dy)/self.a0
    z = (self.zmin+verts[:, 2]*self.dz)/self.a0
    ax.plot_trisurf(x, y, faces, z, alpha=alpha, color=color)

    return ax


# Typing imports
from typing import TYPE_CHECKING, Optional, Any
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# General packages
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import plotly.graph_objects as go
from skimage.measure import marching_cubes

# Local application imports
from comfit.tool import tool_complete_field,\
                        tool_colormap, \
                        tool_set_plot_axis_properties_plotly, \
                        tool_plotly_define_3D_plot_ax, \
                        tool_plotly_colorbar

def plot_complex_field_in_plane_plotly(
        self: 'BaseSystem',
        complex_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> tuple[go.Figure, dict]:
    """Plot the complex field in a plane perpendicular to the given normal vector.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    complex_field : np.ndarray
        The complex field to be plotted.
    normal_vector : np.ndarray, optional
        The normal vector of the plane. Defaults to [0,1,0].
    position : np.ndarray, optional
        The position of the plane. Defaults to the middle of the system.
    \*\*kwargs : Any
        Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns
    -------
    tuple[go.Figure, dict]
        The figure and axes dictionary containing the plot of the complex field.

    Raises
    ------
    Exception
        If the dimension of the system is not 3.
    """

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    complex_field, fig, ax, kwargs = self.plot_prepare(complex_field, field_type = 'complex', **kwargs)

    ax = tool_plotly_define_3D_plot_ax(fig, ax) #Defines sceneN, plot_dimension

    # Extend the field if not a complete array is given
    complex_field = tool_complete_field(self, complex_field)

    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector = [0,1,0]

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    reals = np.real(complex_field)
    imags = np.imag(complex_field)

    # Kewyord arguments
    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    theta_values = theta.ravel()
    rho_values = rho.ravel()

    # Interpolate field at the vertices positions
    # theta_verts = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')
    # rho_verts = sp.interpolate.griddata(points, rho_values, centroids, method='nearest')

    interpolation_method = kwargs.get('interpolation_method', 'nearest') #TODO: This line is inconsistent in naming with
    print("Interpolating points with method: ' ", interpolation_method, "'.")
    print("If this process is slow, consider passing 'interpolation_method='nearest' with the plot_complex_field_in_plane function.") #TODO: this line. (interpolation vs interpolation_method)
    print("That will speed up the process, but the plot may look less smooth.")
    reals_verts = sp.interpolate.griddata(points, reals.ravel(), centroids, method=interpolation_method)
    imags_verts = sp.interpolate.griddata(points, imags.ravel(), centroids, method=interpolation_method)
    print("Interpolation done.")

    theta_verts = np.arctan2(imags_verts, reals_verts)
    rho_verts = np.abs(reals_verts + 1j*imags_verts)

    # Normalize field values for color mapping
    theta_normalized = (theta_verts+np.pi) / (2*np.pi)

    # Map normalized field values to colors
    plt_colormap_object = tool_colormap(kwargs['colormap'], plot_lib='matplotlib')
    colors = plt_colormap_object(theta_normalized)

    # Blend the colors with white according to rho (normalized)
    colors[:,3] = (rho_verts/np.max(rho_verts)).ravel()

    # Create the mesh object
    x_new = (verts[:, 0] * self.dx + self.xmin) / self.a0
    y_new = (verts[:, 1] * self.dy + self.ymin) / self.a0
    z_new = (verts[:, 2] * self.dz + self.zmin) / self.a0

    mesh = go.Mesh3d(
                x=x_new, 
                y=y_new, 
                z=z_new,
                i=faces[:, 0], 
                j=faces[:, 1], 
                k=faces[:, 2],
                facecolor=colors,  # Set color for each face
                showscale=False,
                scene=ax['sceneN']
            )
    fig.add_trace(mesh)

    # Create a colorbar
    # if colorbar:
    #     mappable = plt.cm.ScalarMappable(cmap=tool_colormap_angle())
    #     mappable.set_array([])
    #     mappable.set_clim(-np.pi, np.pi)
    #     cbar = plt.colorbar(mappable, ax=ax, pad=0.2)
    #     cbar.set_ticks(np.array([-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]))
    #     cbar.set_ticklabels([r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])


    if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='angle'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)

    return fig, ax


from typing import Literal, Optional, Union

import numpy as np
import scipy as sp

from comfit.tool.tool_math_functions import tool_multinom

class BaseSystemCalc:
    """Calculation method for the base system class"""
    # CALCULATION FUNCTIONS

    # Calculation of angle fields for vortices of different types
    def calc_angle_field_single_vortex(
        self,
        position: Optional[list] = None,
        charge: int = 1
    ) -> np.ndarray:
        """Calculate the angle field due to a single vortex.
        
        Parameters
        ----------
        position : list, optional
            The position of the vortex. Defaults to None.
        charge : int, optional
            The charge of the vortex. Defaults to 1.
        
        Returns
        -------
        numpy.ndarray
            The angle field calculated for the vortex.
            
        Raises
        ------
        Exception
            If the dimension of the system is not 2.
        """
        if self.dim != 2:
            raise Exception("The dimension of the system must be 2 for a single point vortex.")

        if position is None:
            position = [self.xmid, self.ymid]
        
        theta = charge * np.arctan2(self.y - position[1], self.x - position[0])

        return np.mod(theta + np.pi, 2 * np.pi) - np.pi

    def calc_angle_field_vortex_dipole(
        self,
        dipole_vector: Optional[Union[list[float], tuple[float]]] = None,
        dipole_position: Optional[list[float]] = None
    ) -> np.ndarray:
        """Calculates the angle field for a double vortex system.

        Parameters
        ----------
        dipole_vector : list or tuple, optional
            The dipole vector. If None, defaults to [size_x/3, 0].
        dipole_position : list, optional
            The position of the center of mass of the dipole. If None, defaults to the middle of the grid.
            
        Returns
        -------
        numpy.ndarray
            The calculated angle field for the double vortex system.

        Raises
        ------
        Exception
            If the dimension of the system is not 2.
        """

        if self.dim != 2:
            raise Exception("The dimension of the system must be 2 for a single point vortex.")

        if dipole_vector is None:
            dipole_vector = [self.size_x / 3, 0]

        if dipole_position is None:
            dipole_position = self.rmid
        
        theta = np.zeros(self.dims)
        theta += self.calc_angle_field_single_vortex(dipole_position - np.array(dipole_vector) / 2,
                                                            charge=-1)
        theta += self.calc_angle_field_single_vortex(dipole_position + np.array(dipole_vector) / 2, 
                                                            charge=1)

        # Convert the field to a complex field to make it fit the periodic boundary conditions
        amp = np.exp(1j * theta)

        # Filter the angle field
        width = 0.2 * np.min([self.size_x, self.size_y])
        radius = 0.4 * np.min([self.size_x, self.size_y])

        r2 = self.calc_distance_squared_to_point(self.rmid)
        filter = (1 + np.tanh((radius ** 2 - r2) / width ** 2)) / 2
        amp = amp * filter + (1 - filter)

        theta = np.angle(amp)

        # Roll the field so that the dipole center is at the desired position.
        Rx = round((self.rmid[0] - dipole_position[0]) / self.dx)
        theta = np.roll(theta, -Rx, axis=0)
        Ry = round((self.rmid[1] - dipole_position[1]) / self.dy)
        theta = np.roll(theta, -Ry, axis=1)

        return np.mod(theta + np.pi, 2 * np.pi) - np.pi

    def calc_angle_field_vortex_ring(
            self,
            position: Optional[list[float]] = None,
            radius: Optional[float] = None,
            normal_vector: np.ndarray = [0, 0, 1]
            ) -> np.ndarray:
        """Calculate the angle field for a vortex ring.
        
        Parameters
        ----------
        position : list, optional
            The position of the vortex ring. If None, defaults to the middle of the grid.
        radius : float, optional
            The radius of the vortex ring. If None, defaults to 1/3 of the minimum system size.
        normal_vector : numpy.ndarray, optional
            The normal vector of the vortex ring. Defaults to [0, 0, 1].
            
        Returns
        -------
        numpy.ndarray
            The calculated angle field for the vortex ring.
        """
        if position is None:
            position = self.rmid

        if radius is None:
            radius = np.min([self.size_x, self.size_y, self.size_z]) / 3

        if radius > np.min([self.size_x, self.size_y, self.size_z]) / 3:
            print("Warning: The radius of the suggested vortex ring is large."
                  "This can cause unwanted boundary effects.")

        n = normal_vector / np.linalg.norm(np.array(normal_vector))
        [X, Y, Z] = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        # Add the vortices to the theta-field
        theta = 0
        position = np.array(position)


        m2 = n[0] * (X - position[0]) \
             + n[1] * (Y - position[1]) \
             + n[2] * (Z - position[2])

        m1 = np.sqrt(
            (X - position[0] - m2 * n[0]) ** 2
            + (Y - position[1] - m2 * n[1]) ** 2
            + (Z - position[2] - m2 * n[2]) ** 2
        )

        theta = theta + np.arctan2(m2, m1 + radius)
        theta = theta + np.arctan2(m2, m1 - radius)

        # Convert the field to a complex field to make it fit the periodic boundary conditions
        amp = np.exp(1j * theta)

        # Filter the angle field
        width = 0.2 * np.min([self.size_x, self.size_y, self.size_z])
        radius = 0.4 * np.min([self.size_x, self.size_y, self.size_z])
        # TODO: This radius shares name with the one defining the ring. May cause trouble down the line (Vidar 01.12.23)

        r2 = (self.x.reshape((self.xRes, 1, 1)) - self.xmid) ** 2 \
             + (self.y.reshape((1, self.yRes, 1)) - self.ymid) ** 2 \
             + (self.z.reshape((1, 1, self.zRes)) - self.zmid) ** 2

        filter = (1 + np.tanh((radius ** 2 - r2) / width ** 2)) / 2  # Goes from 1 to 0 continuously
        amp = amp * filter + (1 - filter)

        theta = np.angle(amp)

        # Roll the field so that the ring center is at the desired position.
        Rx = round((self.rmid[0] - position[0]) / self.dx)
        theta = np.roll(theta, -Rx, axis=0)
        Ry = round((self.rmid[1] - position[1]) / self.dy)
        theta = np.roll(theta, -Ry, axis=1)
        Rz = round((self.rmid[2] - position[2]) / self.dz)
        theta = np.roll(theta, -Rz, axis=2)

        return np.mod(theta + np.pi, 2 * np.pi) - np.pi

    def calc_wavenums(
            self, 
            x: np.ndarray
            ) -> np.ndarray:
        """Calculates the wavenumbers corresponding to the input position vectors given by x.

        Parameters
        ----------
        x : numpy.ndarray
            1D array of x-positions.

        Returns
        -------
        numpy.ndarray
            1D array of wavenumbers with all the modes for the given x-array,
            assuming periodicity from x[0] to x[0] over n intervals.

        Examples
        --------
        >>> x = np.array([-10, -5, 0, 5, 10])
        >>> k = instance_of_BaseSystem.calc_wavenums(self, x)
        >>> print(k)
        [ 0.          0.25132741  0.50265482 -0.50265482 -0.25132741]
        """
        n = len(x)

        high = (n - 1) // 2
        low = - (n // 2)

        l = n * (x[1] - x[0])

        k = np.concatenate((np.arange(0, high + 1), np.arange(low, 0))) * 2 * np.pi / l

        return k

    def calc_k2(self):
        """Calculates the squared wavenumber.
        
        Parameters
        ----------
        None
            
        Returns
        -------
        numpy.ndarray
            The squared wavenumber.
        """
        return sum([self.k[i] ** 2 for i in range(len(self.k))])

    def calc_Gaussian_filter_f(self, a0=None):
        """Calculate Gaussian filter in Fourier space.

        This method computes a Gaussian filter in Fourier space using the formula:
        exp(-1/2 * a0^2 * k^2) where k^2 is the squared wavenumber.

        Parameters
        ----------
        a0 : float, optional
            Filter width parameter. If None, uses the instance's a0 attribute.

        Returns
        -------
        numpy.ndarray
            The Gaussian filter in Fourier space.

        Notes
        -----
        The filter is calculated using the squared wavenumber obtained from 
        the calc_k2() method.
        """

        if a0 is None:
            a0 = self.a0

        return np.exp(-1 / 2 * a0 ** 2 * self.calc_k2())

    def calc_determinant_field(self, psi: list[np.ndarray, np.ndarray]) -> np.ndarray:
        """Calculate the determinant transformation of a given field.

        The D-field used in the calculation of the topologicel defect density fields. 
        See the ComFiT documentation for more information:
        https://comfitlib.com/TopologicalDefects/
        
        Parameters
        ----------
        psi : list of numpy.ndarray
            A list of two psi fields.
        
        Returns
        -------
        numpy.ndarray
            The determinant field.
        """
        
        if self.dim == 2:
            if len(psi) == 2:
                psi_f = [sp.fft.fftn(psi[0]), sp.fft.fftn(psi[1])]

                return np.real(
                    sp.fft.ifftn(self.dif[0] * psi_f[0]) * sp.fft.ifftn(self.dif[1] * psi_f[1]) -
                    sp.fft.ifftn(self.dif[1] * psi_f[0]) * sp.fft.ifftn(self.dif[0] * psi_f[1]))
            
        elif self.dim == 3:
            if len(psi) == 2:
                psi_f = [sp.fft.fftn(psi[0]), sp.fft.fftn(psi[1])]

                result = np.array([
                    sp.fft.ifftn(self.dif[1] * psi_f[0]) * sp.fft.ifftn(self.dif[2] * psi_f[1]) -
                    sp.fft.ifftn(self.dif[2] * psi_f[0]) * sp.fft.ifftn(self.dif[1] * psi_f[1]),
                    -sp.fft.ifftn(self.dif[0] * psi_f[0]) * sp.fft.ifftn(self.dif[2] * psi_f[1]) +
                    sp.fft.ifftn(self.dif[2] * psi_f[0]) * sp.fft.ifftn(self.dif[0] * psi_f[1]),
                    sp.fft.ifftn(self.dif[0] * psi_f[0]) * sp.fft.ifftn(self.dif[1] * psi_f[1]) -
                    sp.fft.ifftn(self.dif[1] * psi_f[0]) * sp.fft.ifftn(self.dif[0] * psi_f[1])
                ], dtype='float64')
                #TODO: verify that this is correct to specify as float64. (vidar 26.12.23)

                return np.array(result)

    def calc_defect_density(self, psi: list[np.ndarray, np.ndarray], psi0=1) -> np.ndarray:
        """Calculate the defect density of a given psi field.
        
        Parameters
        ----------
        psi : list of numpy.ndarray
            A list of two psi fields.
        psi0 : float, optional
            The value of psi_0. Default is 1.
            
        Returns
        -------
        numpy.ndarray
            The defect density of the psi field.
        """

        return 1 / (np.pi * psi0 ** 2) * self.calc_determinant_field(psi)

    def calc_defect_density_singular(self, psi: np.ndarray, psi0=1) -> np.ndarray:
        """Calculate the singular defect density for a given psi field.

        Parameters
        ----------
        psi : numpy.ndarray
            The field psi.
        psi0 : float, optional
            The reference value of psi. Default is 1.

        Returns
        -------
        numpy.ndarray
            The defect density for the given psi value.
        """
        return self.calc_defect_density(psi, 1) * self.calc_delta_function(psi, psi0)

    def calc_defect_velocity_field(
        self,
        psi: list[np.ndarray, np.ndarray],
        dt_psi: list[np.ndarray, np.ndarray]
    ) -> list[np.ndarray, np.ndarray]:
        """Calculate the velocity field of the defects in the psi field.

        Parameters
        ----------
        psi : list of numpy.ndarray
            The psi field.
        dt_psi : list of numpy.ndarray
            The time derivative of the psi field.

        Returns
        -------
        list of numpy.ndarray
            The velocity field of the defects.
        """
        if self.dim == 2:
            if len(psi) == 2:
                # Input to exclude region
                threshold = 0.4

                psi_f = [sp.fft.fftn(psi[0]), sp.fft.fftn(psi[1])]

                dx_psi0 = sp.fft.ifftn(self.dif[0] * psi_f[0])
                dy_psi1 = sp.fft.ifftn(self.dif[1] * psi_f[1])
                dx_psi1 = sp.fft.ifftn(self.dif[0] * psi_f[1])
                dy_psi0 = sp.fft.ifftn(self.dif[1] * psi_f[0])

                denominator = np.real(dx_psi0 * dy_psi1 - dx_psi1 * dy_psi0)

                denominator_max = np.max(abs(denominator))
                region_to_ignore = abs(denominator) < threshold * denominator_max

                Vx = -2 * np.real(dt_psi[0] * dy_psi1 - dt_psi[1] * dy_psi0) / denominator
                Vy = -2 * np.real(-dt_psi[0] * dx_psi1 + dt_psi[1] * dx_psi0) / denominator

                #TODO: check if this factor of 2 is actually supposed to be there (Vidar 05.12.23)

                # print(region_to_ignore.shape)
                # print(Vx.shape)
                # print(Vy.shape)

                Vx[region_to_ignore] = 0
                Vy[region_to_ignore] = 0

                return [Vx, Vy]

        elif self.dim == 3:
            if len(psi) == 2:
                # Input to exclude region
                threshold = 0.4

                psi_f = [sp.fft.fftn(psi[0]), sp.fft.fftn(psi[1])]

                dx_psi0 = sp.fft.ifftn(self.dif[0] * psi_f[0])
                dy_psi0 = sp.fft.ifftn(self.dif[1] * psi_f[0])
                dz_psi0 = sp.fft.ifftn(self.dif[2] * psi_f[0])

                dx_psi1 = sp.fft.ifftn(self.dif[0] * psi_f[1])
                dy_psi1 = sp.fft.ifftn(self.dif[1] * psi_f[1])
                dz_psi1 = sp.fft.ifftn(self.dif[2] * psi_f[1])


                denominator = 2 * (dx_psi1 ** 2 * (dy_psi0 ** 2 + dz_psi0 ** 2) + (
                            dy_psi1 * dz_psi0 - dy_psi0 * dz_psi1) ** 2 -
                                   2 * dx_psi0 * dx_psi1 * (dy_psi0 * dy_psi1 + dz_psi0 * dz_psi1) +
                                   dx_psi0 ** 2 * (dy_psi1 ** 2 + dz_psi1 ** 2))

                Vx = -2 * np.real((
                          (-dx_psi1 * (dy_psi0 * dy_psi1 + dz_psi0 * dz_psi1) +
                           dx_psi0 * (dy_psi1 ** 2 + dz_psi1 ** 2)) * dt_psi[0] +
                          (dx_psi1 * (dy_psi0 ** 2 + dz_psi0 ** 2) -
                           dx_psi0 * (dy_psi0 * dy_psi1 + dz_psi0 * dz_psi1)) * dt_psi[1]
                                  )/denominator)
                Vy = -2 * np.real((
                          (dx_psi1 ** 2 * dy_psi0 - dx_psi0 * dx_psi1 * dy_psi1 +
                           dz_psi1 * (-dy_psi1 * dz_psi0 + dy_psi0 * dz_psi1)) * dt_psi[0] +
                          (-dx_psi0 * dx_psi1 * dy_psi0 + dx_psi0 ** 2 * dy_psi1 +
                           dz_psi0 * (dy_psi1 * dz_psi0 - dy_psi0 * dz_psi1)) * dt_psi[1]
                                  ) / denominator)

                Vz = -2 * np.real((
                          ((dx_psi1 ** 2 + dy_psi1 ** 2) * dz_psi0 - (
                                      dx_psi0 * dx_psi1 + dy_psi0 * dy_psi1) * dz_psi1) * dt_psi[0] +
                          (-(dx_psi0 * dx_psi1 + dy_psi0 * dy_psi1) * dz_psi0 + (
                                      dx_psi0 ** 2 + dy_psi0 ** 2) * dz_psi1) * dt_psi[1]
                                ) / denominator)

                denominator_max = np.max(abs(denominator))
                region_to_ignore = abs(denominator) < threshold * denominator_max

                Vx[region_to_ignore] = 0
                Vy[region_to_ignore] = 0
                Vz[region_to_ignore] = 0

                return [Vx, Vy, Vz]

    def calc_defect_current_density(self, psi, dt_psi, psi_0=0):
        """Calculates the conserved current of the superfluid density.
        
        Parameters
        ----------
        psi : numpy.ndarray
            The vector field that we find the density of.
        dt_psi : numpy.ndarray
            The time derivative of psi.
        psi_0 : float or numpy.ndarray, optional
            The equilibrium state. Default is 0.
        
        Returns
        -------
        list of numpy.ndarray
            Components of the conserved current.
        """
        if self.dim == 2:
            if len(psi) == 2:
                psi_f = [sp.fft.fftn(psi[0]), sp.fft.fftn(psi[1])]

                dx_psi0 = sp.fft.ifftn(self.dif[0] * psi_f[0])
                dy_psi1 = sp.fft.ifftn(self.dif[1] * psi_f[1])
                dx_psi1 = sp.fft.ifftn(self.dif[0] * psi_f[1])
                dy_psi0 = sp.fft.ifftn(self.dif[1] * psi_f[0])

                Jx = -  np.real(dt_psi[0] * dy_psi1 - dt_psi[1] * dy_psi0) / (psi_0 * np.pi)
                Jy = - np.real(-dt_psi[0] * dx_psi1 + dt_psi[1] * dx_psi0) / (psi_0 * np.pi)

                return [Jx, Jy]

    def calc_delta_function(self, psi: list[np.ndarray, np.ndarray], psi0=1):
        """Calculate the delta function for a given wavefunction.

        Parameters
        ----------
        psi : list of numpy.ndarray
            The wavefunction.
        psi0 : float, optional
            The width of the wavefunction. Default is 1.

        Returns
        -------
        numpy.ndarray
            The value of the delta function.
        """
        width = psi0 / 2
        n = len(psi)
        if self.dim == 2:
            if n == 2:
                psi2 = psi[0] ** 2 + psi[1] ** 2
                return 1 / (2 * np.pi * width ** 2) * np.exp(-psi2 / (2 * width ** 2))

    def calc_region_interval(self, a: float, b: float) -> np.ndarray:
        """Calculate a boolean array indicating whether a point is within an interval.
        
        Parameters
        ----------
        a : float
            The lower bound of the interval.
        b : float
            The upper bound of the interval.
        
        Returns
        -------
        numpy.ndarray
            A boolean array indicating whether a point is within the interval.
            
        Raises
        ------
        Exception
            If the lower bound is greater than the upper bound or if the dimension of the system is not 1.
        """
        if not (a <= b):
            raise Exception("The lower bound must be less than or equal to the upper bound.")
        
        if not (self.dim == 1):
            raise Exception("This function is only valid for 1D systems.")
        
        return (a <= self.x) & (self.x <= b)
        

    def calc_region_disk(self, position: list[float], radius: float) -> np.ndarray:
        """Calculates a boolean array indicating whether a point is within a disk of a given radius.
        
        Parameters
        ----------
        position : list[float]
            The position of the disk
        radius : float
            The radius of the disk
        
        Returns
        -------
        numpy.ndarray
            A boolean array indicating whether a point is within the disk.
            
        Raises
        ------
        Exception
            If the dimension of the system is not 2.
        """
        if self.dim == 2:
            return self.calc_distance_squared_to_point(position) <= radius ** 2

        else:
            raise Exception("Not valid for other dimensions.")

    def calc_region_ball(self, 
            position: list[float], 
            radius: float) -> np.ndarray:
        """Calculate a boolean array indicating whether a point is within a ball of a given radius.
        
        Parameters
        ----------
        position : list[float]
            The position of the ball
        radius : float
            The radius of the ball
        
        Returns
        -------
        numpy.ndarray
            A boolean array indicating whether a point is within the ball.
            
        Raises
        ------
        Exception
            If the dimension of the system is not 3.
        """
        if self.dim == 3:
            return self.calc_distance_squared_to_point(position) <= radius ** 2

        else:
            raise Exception("Not valid for other dimensions.")

    def calc_Gaussian(self, 
            position=None, 
            width=None, 
            top=None,
            value=None):
        """Calculate the Gaussian function.
        
        Parameters
        ----------
        position : list or array, optional
            The center position of the Gaussian. If None, uses the middle of the grid.
        width : float, optional
            The width of the Gaussian. If None, uses the system's a0 parameter.
        top : float, optional
            The maximum value of the Gaussian function.
        value : float, optional
            The integrated value of the Gaussian function.
            If neither top nor value is provided, the function will be normalized to 1.
        
        Returns
        -------
        numpy.ndarray
            The Gaussian function evaluated on the grid.
        
        Notes
        -----
        If top is provided, it scales the height of the Gaussian.
        If value is provided, it scales the integral of the Gaussian.
        """

        if position is None:
            position = self.rmid
        if width is None:
            width = self.a0

        r2 = self.calc_distance_squared_to_point(position)

        if top is not None:
            return top * np.exp(-r2 / (2 * width ** 2))

        if value is None:
            value =1

        return value*(2*np.pi*width**2)**(-self.dim/2)*np.exp(-r2/(2*width**2))

    def calc_distance_squared_to_point(self, position: Union[float, list[float]]) -> np.ndarray:
        """Calculate the squared distance from each point in the grid to a specified position.

        Parameters
        ----------
        position : Union[float, list[float]]
            The target position. Should be a single float for 1D systems or a list of floats for higher dimensions.

        Returns
        -------
        np.ndarray
            An array of the same dimensionality as the system containing the squared distance 
            from each grid point to the specified position. The shape matches the grid resolution.

        Notes
        -----
        This method accounts for periodic boundary conditions by calculating distances in each
        dimension considering the original position, and positions shifted by +/- the system size.
        It then takes the minimum of these distances to find the shortest path under periodicity.
        """

        if self.dim == 1:
            position = [position]
        
        rx2m = (self.x - position[0] - self.size_x) ** 2
        rx2 = (self.x - position[0]) ** 2
        rx2p = (self.x - position[0] + self.size_x) ** 2

        r2 = np.min(np.stack((rx2m, rx2, rx2p)), axis=0).reshape((self.xRes))

        if self.dim > 1:
            r2 = r2.reshape((self.xRes, 1))

            ry2m = (self.y - position[1] - self.size_y) ** 2
            ry2 = (self.y - position[1]) ** 2
            ry2p = (self.y - position[1] + self.size_y) ** 2
            ry2 = np.min(np.stack((ry2m, ry2, ry2p)), axis=0).reshape((1, self.yRes))

            r2 = r2 + ry2
        
        if self.dim > 2:
            r2 = r2.reshape((self.xRes, self.yRes, 1))
            
            rz2m = (self.z - position[2] - self.size_z) ** 2
            rz2 = (self.z - position[2]) ** 2
            rz2p = (self.z - position[2] + self.size_z) ** 2
            rz2 = np.min(np.stack((rz2m, rz2, rz2p)), axis=0).reshape((1, 1, self.zRes))

            r2 = r2 + rz2

        return r2

    def calc_region_cylinder(
        self,
        position: list[float],
        radius: float,
        normal_vector: np.ndarray,
        height: float
    ) -> np.ndarray:
        """Calculates a boolean array indicating whether a point is within a cylinder of a given radius and height.
        
        Parameters
        ----------
        position : list[float]
            The position of the cylinder
        radius : float
            The radius of the cylinder
        normal_vector : np.ndarray
            The normal vector of the cylinder
        height : float
            The height of the cylinder
        
        Returns
        -------
        np.ndarray
            A boolean array indicating whether a point is within the cylinder.
            
        Raises
        ------
        Exception
            If the dimension of the system is not 3.
        """

        if self.dim == 3:
            t = normal_vector / np.linalg.norm(np.array(normal_vector))

            rx = (self.x - position[0])
            # Points that are further away than half the size of the box are moved to the other side
            rx[rx > self.size_x / 2] = rx[rx > self.size_x / 2] - self.size_x
            rx[rx < -self.size_x / 2] = rx[rx < -self.size_x / 2] + self.size_x

            ry = (self.y - position[1])
            ry[ry > self.size_y / 2] = ry[ry > self.size_y / 2] - self.size_y
            ry[ry < -self.size_y / 2] = ry[ry < -self.size_y / 2] + self.size_y

            rz = (self.z - position[2])
            rz[rz > self.size_z / 2] = rz[rz > self.size_z / 2] - self.size_z
            rz[rz < -self.size_z / 2] = rz[rz < -self.size_z / 2] + self.size_z

            zt = rx * t[0] + ry * t[1] + rz * t[2]

            # Project to perpendicular plane vector
            Rt2 = (rx - zt * t[0]) ** 2 + (ry - zt * t[1]) ** 2 + (rz - zt * t[2]) ** 2

            return (zt ** 2 <= height ** 2) & (Rt2 <= radius ** 2)

        else:
            raise Exception("Not valid for other dimensions.")

    def calc_integrate_field(self, field: np.ndarray, region: Optional[int] = None) -> float:
        """Calculates the integrated field value within a specified region.
        
        Parameters
        ----------
        field : numpy.ndarray
            The field array to integrate.
        region : Optional[int], default=None
            If provided, specifies the region to integrate over.
            If None, integrates over the entire field.
        
        Returns
        -------
        float
            The integrated field value within the region.
            
        Raises
        ------
        Exception
            If the dimension of the field is not compatible.
        """

        if region is None:
            return np.sum(field) * self.dV
        else:
            return np.sum(field[region]) * self.dV

    def calc_integrating_factors_f_and_solver(self, omega_f, method: Literal["ETD2RK", "ETD4RK"]) -> tuple:
        """Calculates the integrating factors and the solver for the evolution equation.
        
        Parameters
        ----------
        omega_f : numpy.ndarray
            The value of omega_f in Fourier space.
        method : {'ETD2RK', 'ETD4RK'}
            The method used for evolution.
        
        Returns
        -------
        tuple
            A tuple containing: integrating_factors_f : list (The integrating 
            factors for the selected method), solver : function (The 
            corresponding solver function for the evolution equation).
                
        Raises
        ------
        Exception
            If the specified method is not implemented.
        """
        if method == 'ETD2RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD2RK(omega_f)
            solver = self.evolve_ETD2RK_loop

        elif method == 'ETD4RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD4RK(omega_f)
            solver = self.evolve_ETD4RK_loop
        else:
            raise Exception("This method is not implemented.")

        return integrating_factors_f, solver

    def calc_advect_field(
        self,
        field: np.ndarray,
        u: np.ndarray,
        field_f: Optional[np.ndarray] = None,
        order: int = 3
    ) -> np.ndarray:
        """Advects field according to the provided displacement field u.
        
        Uses a Taylor expansion up to specified order to compute the advection.
        
        Parameters
        ----------
        field : np.ndarray
            The field to be advected.
        u : np.ndarray
            Displacement field.
        field_f : np.ndarray, optional
            Fourier transform of the field. If None, it will be calculated.
        order : int, default=3
            Order of the Taylor expansion. Must be less than or equal to 3.
            
        Returns
        -------
        np.ndarray
            The field after advection by u.
            
        Raises
        ------
        ValueError
            If the order of the Taylor expansion is greater than 3.
        """

        if order > 3:
            raise ValueError("The order of the taylor expansion must be less than or equal to 3.")

        if field_f is None:
            field_f = sp.fft.fftn(field)

        if order > 0:
            for i in range(self.dim):
                # Calculate the derivative
                difield = sp.fft.ifftn(self.dif[i]*field_f)
                # Advect the PFC
                field = field - u[i]*difield

        if order > 1:
            for i in range(self.dim):
                for j in range(i, self.dim):
                    # Calculate the derivative
                    dijfield = sp.fft.ifftn(self.dif[i]*self.dif[j]*field_f)
                    # Advect the PFC
                    field = field + tool_multinom(i,j)*u[i]*u[j]*dijfield
        
        if order > 2:
            for i in range(self.dim):
                for j in range(i, self.dim):
                    for k in range(j, self.dim):
                        # Calculate the derivative
                        dijkfield = sp.fft.ifftn(self.dif[i]*self.dif[j]*self.dif[k]*field_f)
                        # Advect the PFC
                        field = field - tool_multinom(i,j,k)*u[i]*u[j]*u[k]*dijkfield

        return field

    # def calc_advect_field_alot(
    #     self,
    #     field: np.ndarray,
    #     u: np.ndarray,
    #     field_f: Optional[np.ndarray] = None,
    #     order: int = 3
    # ) -> np.ndarray:

    # if self.dim == 2:
        
    #     Ux = u[0]/self.dx
    #     Uy = u[1]/self.dy



    def calc_evolution_integrating_factors_ETD2RK(self, omega_f: np.ndarray, tol: float = 10 ** (-5)) -> list:
        """Calculates integrating factors for ETD2RK.
        
        Parameters
        ----------
        omega_f : numpy.ndarray
            The value of omega_f in Fourier space.
        tol : float, optional
            Tolerance for when to expand the integrating factors that divide by omega.
            Default is 10^-5.
        
        Returns
        -------
        list
            The list of integrating factors for the ETD2RK method.
        """
        integrating_factors_f = [0, 0, 0]

        integrating_factors_f[0] = np.exp(omega_f * self.dt)
        If1 = integrating_factors_f[0]

        integrating_factors_f[1] = np.divide(If1 - 1, omega_f, 
                                    out=np.full_like(omega_f, self.dt), #Small omega_f * self.dt values are set to low value limit
                                    where=np.abs(omega_f) * self.dt >= tol)

        integrating_factors_f[2] = np.divide(If1 - 1 - omega_f * self.dt, self.dt * omega_f ** 2,
                                    out=np.full_like(omega_f, self.dt / 2), #Small omega_f * self.dt values are set to low value limit
                                    where=np.abs(omega_f) * self.dt >= tol)

        return integrating_factors_f

    def calc_evolution_integrating_factors_ETD4RK(self, omega_f: np.ndarray, tol: float = 10 ** (-5)) -> list:
        """Calculate the evolution integrating factors using the ETDRK4 method.
            
        Parameters
        ----------
        omega_f : numpy.ndarray
            The value of omega_f in Fourier space.
        tol : float, optional
            Tolerance for when to expand the integrating factors that divide by omega.
            Default is 10^-5.
             
        Returns
        -------
        list
            The list of integrating factors for the ETD4RK method.
        """
        integrating_factors_f = [0, 0, 0, 0, 0, 0]

        integrating_factors_f[0] = np.exp(omega_f * self.dt / 2)
        If1 = integrating_factors_f[0]

        integrating_factors_f[1] = np.divide(If1 - 1, omega_f,
                        out=np.full_like(omega_f, self.dt / 2), #Small omega_f * self.dt values are set to low value limit
                        where=np.abs(omega_f) * self.dt >= tol)

        integrating_factors_f[2] = np.exp(omega_f * self.dt)
        If2 = integrating_factors_f[2]

        integrating_factors_f[3] = np.divide(
            -4 - omega_f * self.dt + If2 * (4 - 3 * omega_f * self.dt + omega_f ** 2 * self.dt ** 2),
            omega_f ** 3 * self.dt ** 2,
            out=np.full_like(omega_f, self.dt / 6), #Small omega_f * self.dt values are set to low value limit
            where=np.abs(omega_f) * self.dt >= tol)

        integrating_factors_f[4] = np.divide(
            2 * (2 + omega_f * self.dt + If2 * (-2 + omega_f * self.dt)),
            omega_f ** 3 * self.dt ** 2,
            out=np.full_like(omega_f, self.dt / 3), #Small omega_f * self.dt values are set to low value limit
            where=np.abs(omega_f) * self.dt >= tol)

        integrating_factors_f[5] = np.divide(
            -4 - 3 * omega_f * self.dt - omega_f ** 2 * self.dt ** 2 + If2 * (4 - omega_f * self.dt),
            omega_f ** 3 * self.dt ** 2,
            out=np.full_like(omega_f, self.dt / 6), #Small omega_f * self.dt values are set to low value limit
            where=np.abs(omega_f) * self.dt >= tol)
        
        return integrating_factors_f

    def calc_defect_nodes(
        self,
        defect_density: np.ndarray,
        charge_tolerance: Optional[float] = None,
        integration_radius: Optional[float] = None
    ):
        """Calculate the positions and charges of defect nodes based on the defect density.
        
        Parameters
        ----------
        defect_density : numpy.ndarray
            The defect density field. A positive scalar field to be integrated.
        charge_tolerance : float, optional
            The minimum charge value to consider as a defect. If None, uses default values based on dimension.
        integration_radius : float, optional
            The radius to use for integrating around potential defects. If None, uses default values based on dimension.
        
        Returns
        -------
        list
            A list of dictionaries representing the defect nodes. Each dictionary contains:
            - 'position_index': The position index of the defect node in the defect density array.
            - 'position': The physical position of the defect node.
            
        Raises
        ------
        Exception
            If the defect density is not a positive scalar field.
        """

        if not (defect_density>=0).all():
            raise Exception("Defect density must be a positive real scalar field.")

        defect_nodes = []

        if self.dim == 2:
            if charge_tolerance is None:
                charge_tolerance = 0.2
            if integration_radius is None:
                integration_radius = self.a0
            
            #Auxiliary functions
            def calc_region(position_index,radius):
                return self.calc_region_disk(position=[
                    self.x.flatten()[position_index[0]],
                    self.y.flatten()[position_index[1]]], 
                    radius=radius)

            def calc_position_from_region(defect_density,region_to_integrate,position_index):
                # Calculate the field to integrate
                field_to_integrate = region_to_integrate * defect_density
                # Normalize the field to integrate
                field_to_integrate = field_to_integrate / np.sum(field_to_integrate)

                # Roll the field so that the defect node is at the center
                Rx = round((self.x.flatten()[position_index[0]] - self.xmid) / self.dx)
                field_to_integrate = np.roll(field_to_integrate, -Rx, axis=0)
                Ry = round((self.y.flatten()[position_index[1]] - self.ymid) / self.dy)
                field_to_integrate = np.roll(field_to_integrate, -Ry, axis=1)
                
                x = np.sum(field_to_integrate * self.x) 
                y = np.sum(field_to_integrate * self.y) 

                x = x + Rx*self.dx
                y = y + Ry*self.dy

                return [x,y]

        elif self.dim == 3:
            if charge_tolerance is None:
                charge_tolerance = 0.5*self.a0**2
                # print("charge tolerance",charge_tolerance)
            if integration_radius is None:
                integration_radius = 2*self.a0

            #Auxiliary functions
            def calc_region(position_index,radius):
                return self.calc_region_ball(position=[
                        self.x.flatten()[position_index[0]],
                        self.y.flatten()[position_index[1]],
                        self.z.flatten()[position_index[2]]], 
                        radius=radius)
            
            def calc_position_from_region(defect_density,region_to_integrate,position_index):
                # Calculate the field to integrate
                field_to_integrate = region_to_integrate * defect_density
                # Normalize the field to integrate
                field_to_integrate = field_to_integrate / np.sum(field_to_integrate)

                # Roll the field so that the defect node is at the center
                Rx = round((self.x.flatten()[position_index[0]] - self.xmid) / self.dx)
                field_to_integrate = np.roll(field_to_integrate, -Rx, axis=0)
                Ry = round((self.y.flatten()[position_index[1]] - self.ymid) / self.dy)
                field_to_integrate = np.roll(field_to_integrate, -Ry, axis=1)
                Rz = round((self.z.flatten()[position_index[2]] - self.zmid) / self.dz)
                field_to_integrate = np.roll(field_to_integrate, -Rz, axis=2)

                x = np.sum(field_to_integrate * self.x)
                y = np.sum(field_to_integrate * self.y)
                z = np.sum(field_to_integrate * self.z)

                x = x + Rx*self.dx
                y = y + Ry*self.dy
                z = z + Rz*self.dz

                return [x,y,z]

        #Region to search for defect nodes
        region_to_search = np.ones(self.dims)

        # Calculate the point where defect density is largest
        position_index = np.unravel_index(np.argmax(defect_density*region_to_search), defect_density.shape)

        # Integrate the defect density around this point (i.e. in a disk/ball around)
        region_to_integrate = calc_region(position_index,
                                        radius=integration_radius)

        charge = self.calc_integrate_field(defect_density, region_to_integrate)

        # print("Charge: ", charge)
        # print("Charge tolerance: ", charge_tolerance)

        while charge > charge_tolerance:
            defect_node = {}
            defect_node['position_index'] = position_index

            defect_node['position'] = calc_position_from_region(defect_density,region_to_integrate,position_index)

            # print("Defect node position: ", defect_node['position'])
            defect_nodes.append(defect_node)

            region_to_exclude_from_search = calc_region(position_index,
                                        radius=2*integration_radius)
        
            region_to_search[region_to_exclude_from_search] = 0

            position_index = np.unravel_index(np.argmax(defect_density*region_to_search), defect_density.shape)

            region_to_integrate = calc_region(position_index,
                                        radius=integration_radius)
            
            charge = self.calc_integrate_field(defect_density, region_to_integrate)

        return defect_nodes


from comfit.phase_field_crystal.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint
from comfit.tool.tool_print_in_color import tool_print_in_color


class PhaseFieldCrystal3DBodyCenteredCubic(PhaseFieldCrystal):
    def __init__(self, nx, ny, nz, **kwargs):
        """
        Initializes a phase field crystal system in 3D with a body centered cubic crystal structure.

        Parameters
        ----------
        nx : int
            The number of unit cells in the x direction.
        ny : int
            The number of unit cells in the y direction.
        nz : int
            The number of unit cells in the z direction.
        \*\*kwargs : dict
            Additional arguments to set as attributes.

        Returns
        -------
        None
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal3DBodyCenteredCubic'
        self.dim = 3
        self.nx = nx
        self.ny = ny
        self.nz = nz

        # Default simulation parameters
        self.micro_resolution = [7, 7, 7]
        self.psi0 = -0.325
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]
        self.zRes = nz * self.micro_resolution[2]

        a0 = 2 * np.pi * np.sqrt(2)
        self.a = a0 / 2 * np.array([[-1, 1, 1],
                                    [1, -1, 1],
                                    [1, 1, -1],
                                    [1, 1, 1]])

        self.q = np.array([[0, 1, 1],
                           [1, 0, 1],
                           [1, 1, 0],
                           [0, -1, 1],
                           [-1, 0, 1],
                           [-1, 1, 0]]) / np.sqrt(2)

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 6
        self.number_of_primary_reciprocal_lattice_modes = 6

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = a0 / self.micro_resolution[1]
        self.dz = a0 / self.micro_resolution[2]

        self.type_of_evolution = kwargs.get('type_of_evolution', 'conserved')
        self.A_proto = self.calc_proto_amplitudes_conserved()

        # Set the elastic constants
        self.el_lambda_proto = 4 * self.A_proto ** 2
        self.el_mu_proto = 4 * self.A_proto ** 2
        self.el_gamma_proto = - 4*self.A_proto**2

        bool_is_for_properties_calculation = kwargs.get('for_properties_calculation', False)
        if not bool_is_for_properties_calculation:
            tool_print_in_color('Initiating a 3D bcc PFC model.', 'green')
            kwargs.pop('for_properties_calculation', None)
            pfc = PhaseFieldCrystal3DBodyCenteredCubic(1,1,1,for_properties_calculation=True, type_of_evolution=self.type_of_evolution, **kwargs)
            final_strain, self.psi0, self.A, self.el_lambda, self.el_mu, self.el_gamma = pfc.calc_strained_amplitudes()  
        else:
            self.A = self.A_proto
            self.el_lambda = self.el_lambda_proto
            self.el_mu = self.el_mu_proto
            self.el_gamma = self.el_gamma_proto


        self.eta0 = np.array([self.A, self.A, self.A, self.A, self.A, self.A])

        # Initialize the BaseSystem
        kwargs['xRes'] = self.xRes
        kwargs['yRes'] = self.yRes
        kwargs['zRes'] = self.zRes
        kwargs['dx'] = self.dx
        kwargs['dy'] = self.dy
        kwargs['dz'] = self.dz
        kwargs['dt'] = self.dt
        super().__init__(self.dim, **kwargs)

        # Set the a0
        self.a0 = a0
        self.bool_has_defined_length_scale = True

        if not bool_is_for_properties_calculation:
            self.conf_apply_distortion([[final_strain,0,0],[0,final_strain,0],[0,0,final_strain]], update_q_and_a_vectors=True)
            self.a0 = self.a0 * (1+final_strain)

    def calc_proto_amplitudes_conserved(self):
        """
        Calculates the proto-amplitudes for the system.

        Returns
        -------
        float
            The proto-amplitudes for the system.
        """
        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        # Testing which of the three amplitudes give the lowest free energy
        A = 0
        A_tmp = 1 / (45 * v) * (-2 * t - 6 * v * psi0 + np.sqrt(4 * t**2 - 45 * r * v - 66 * t * v * psi0 - 99 * v**2 * psi0**2))
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        A_tmp = 1 / (45 * v) * (-2 * t - 6 * v * psi0 - np.sqrt(4 * t**2 - 45 * r * v - 66 * t * v * psi0 - 99 * v**2 * psi0**2))
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        return A

    def calc_free_energy_from_proto_amplitudes(self, psi0, A):
        """
        Calculates the free energy of the phase-field crystal from the proto amplitudes.

        Parameters
        ----------
        psi0 : float
            The average value of psi.
        A : float
            The proto-amplitude.

        Returns
        -------
        float
            The free energy of the phase-field crystal.
        """
        r = self.r
        t = self.t
        v = self.v

        return 4 * np.sqrt(2) * np.pi**3 / 3 * (1620 * A**4 * v + 192 * A**3 * (t + 3 * v * psi0) + psi0**2 * (6 + 6 * r + 4 * t * psi0 + 3 * v * psi0**2) + 72 * A**2 * (r + psi0 * (2 * t + 3 * v * psi0)))

    def calc_L_f(self):
        """Calculate the L operator in Fourier space.

        Returns
        -------
        numpy.ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return 1 - k2

    def calc_L_sum_f(self):
        """Calculate the sum of the L operators in Fourier space. Needed for stress calculation functions.

        Returns
        -------
        int
            The L operator in Fourier space.
        """
        return 1


from typing import Dict, Any

# General packages
import plotly.graph_objects as go

# Local packages
from comfit.tool import tool_plotly_find_next_xN

def tool_plotly_define_2D_plot_ax(
        fig: go.Figure, 
        ax: Dict[str, Any]
        ) -> Dict[str, Any]:
    """Define x-axis, y-axis and plot dimension parameters for 2D Plotly plots.

    This function sets up axis naming conventions and validates plot dimensions
    for 2D plots in Plotly. It determines the next available x-axis identifier
    and corresponding y-axis.
    
    Parameters
    ----------
    fig : plotly.graph_objects.Figure
        The Plotly figure object to define axes for.
    ax : Dict[str, Any]
        Dictionary containing axis parameters. May include:
        xN : str (optional, X-axis identifier, e.g., 'x', 'x2', 'x3'),
        yN : str (optional, Y-axis identifier, e.g., 'y', 'y2', 'y3'),
        plot_dimension : int (optional, plot dimension, must be 2 for 2D plots).
    
    Returns
    -------
    Dict[str, Any]
        Updated dictionary containing:
        xN : str (X-axis identifier),
        yN : str (Y-axis identifier),
        plot_dimension : int (Plot dimension, always 2),
        xaxisN : str (Full x-axis reference, e.g., 'xaxis', 'xaxis2'),
        yaxisN : str (Full y-axis reference, e.g. 'yaxis', 'yaxis2').

    Raises
    ------
    ValueError
        If plot_dimension is not 2.

    Notes
    -----
    The function automatically determines the next available x-axis identifier
    if not provided in the input dictionary.
    """

    #Plot nature
    ax['xN'] = ax.get('xN', tool_plotly_find_next_xN(fig))
    ax['yN'] = 'y' if ax['xN'] == 'x' else f'y{ax["xN"][1:]}'

    # Plot dimension
    ax['plot_dimension'] = ax.get('plot_dimension', 2)
    if ax['plot_dimension'] != 2:
        raise ValueError(f"Plot dimension must be 2 for 2D plots. Got {ax['plot_dimension']}")

    # Axis number
    axis_number = 1 if ax['xN'] == 'x' else int(ax['xN'][1:])  # extracts everything after 'x' and converts to integer
    
    ax['xaxisN'] = f'xaxis{axis_number}'
    ax['yaxisN'] = f'yaxis{axis_number}'

    return ax


# Standard library imports
from typing import Tuple, Any

# Third-party library imports
import matplotlib.pyplot as plt
import numpy as np


def plot_subplots_matplotlib(
    number_of_rows: int,
    number_of_columns: int
    ) -> Tuple[plt.Figure, plt.Axes]:
    """Create a figure with a grid of subplots.

    Parameters
    ----------
    number_of_rows : int
        Number of rows in the subplot grid.
    number_of_columns : int
        Number of columns in the subplot grid.

    Returns
    -------
    Tuple[plt.Figure, plt.Axes]
        Figure object and array of Axes objects representing the subplots.
    """
    fig, axs = plt.subplots(number_of_rows, number_of_columns)
    return fig, axs


# General packages
import plotly.graph_objects as go


def tool_plotly_find_next_sceneN(
        fig : go.Figure
        ) -> str:
    """Find the next available scene name in a Plotly figure.

    Parameters
    ----------
    fig : plotly.graph_objects.Figure
        The Plotly figure to analyze.

    Returns
    -------
    str
        The next available scene name. Returns 'scene' if no scenes are used,
        otherwise returns 'sceneN' where N is one more than the current highest scene number.
    """
    
    used_scene = set()
    
    # Check all traces in the figure
    for trace in fig.data:
        if hasattr(trace, 'scene'):
            used_scene.add(trace.scene)
    
    return 'scene' if not used_scene else f'scene{len(used_scene)+1}'


if __name__ == "__main__":
    import plotly.graph_objects as go
    
    fig = go.Figure()
    fig.add_trace(go.Scatter3d(x=[1, 2, 3], y=[4, 5, 6], z=[7, 8, 9], scene='scene'))
    fig.add_trace(go.Scatter3d(x=[2, 3, 4], y=[5, 6, 7], z=[8, 9, 10], scene='scene2'))

    print(tool_plotly_find_next_sceneN(fig))
    print('Expected: scene3')

    fig = go.Figure()
    print(tool_plotly_find_next_sceneN(fig))
    print('Expected: scene')


from typing import Dict, Tuple, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
    
# Third-party library imports
import plotly.graph_objects as go

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_set_plot_axis_properties_matplotlib,
    tool_set_plot_axis_properties_plotly,
    tool_colormap,
    tool_plotly_find_next_xN,
    tool_plotly_define_2D_plot_ax,
    tool_plotly_define_3D_plot_ax,
    tool_plotly_colorbar
)

def plot_field_plotly(
    self: 'BaseSystem',
    field: np.ndarray,
    **kwargs: Any
) -> Tuple[go.Figure, Dict]:
    """Plot the given (real) field using Plotly.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    field : np.ndarray
        The field to be plotted.
    \*\*kwargs : Any
        Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ 
        for a full list of keyword arguments.

    Returns
    -------
    Tuple[go.Figure, Dict]
        A tuple containing the Plotly figure and axes dictionary.
    """

    field, fig, ax, kwargs = self.plot_prepare(field, field_type = 'real', **kwargs)

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        ax = tool_plotly_define_2D_plot_ax(fig, ax) #Defines xN, yN and plot_dimension

        if not kwargs['field_is_nan']:
            trace = go.Scatter(
                x=self.x/self.a0,
                y=field,
                mode='lines',
                name='',
                hovertemplate='x: %{x:.1f} a₀<br>field: %{y:.1f}',
                xaxis=ax['xN'],
                yaxis=ax['yN'],
                showlegend=False
            )
            fig.add_trace(trace)

    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    if self.dim == 2:
        
        ax = tool_plotly_define_2D_plot_ax(fig, ax) #Defines xN, yN and plot_dimension

        X = kwargs.get('X', None)
        Y = kwargs.get('Y', None)

        if X is None or Y is None:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')
            
        opacity = kwargs.get('opacity', 1)

        if not kwargs['field_is_nan']:
            # Trace
            trace = go.Heatmap(
                x=X.flatten()/self.a0,
                y=Y.flatten()/self.a0,
                z=field.flatten(),
                zmin=ax['vmin'],
                zmax=ax['vmax'],
                zsmooth='best',
                hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br> field: %{z:.2f}',
                name='',
                opacity=opacity,
                colorscale=kwargs['colormap_object'],
                showscale=False,
                xaxis=ax['xN'],
                yaxis=ax['yN']
            )

            fig.add_trace(trace)
        

    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################
    elif self.dim == 3:

        ax = tool_plotly_define_3D_plot_ax(fig, ax) #Defines sceneN, plot_dimension

        # Keyword arguments particular to the 3D case
        number_of_layers = kwargs.get('number_of_layers', 1)
        alpha = kwargs.get('alpha', 0.5)

        if 'layer_values' in kwargs:
            layer_values = np.concatenate([[-np.inf], kwargs['layer_values'], [np.inf]])
        else: 
            layer_values = np.linspace(ax['vmin'], ax['vmax'], number_of_layers + 2)


        #Plotting the layers
        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        if not kwargs['field_is_nan']:
            for layer_value in layer_values[1:-1]:
                
                trace = go.Isosurface(
                    x=X.flatten()/self.a0,
                    y=Y.flatten()/self.a0,
                    z=Z.flatten()/self.a0,
                    value = field.flatten(),
                    isomin = layer_value,
                    isomax = layer_value,
                    cmin = ax['vmin'],
                    cmax = ax['vmax'],
                    colorscale = kwargs['colormap_object'],
                    showscale=False,
                    surface=dict(count=3),  # Ensuring only one surface is shown
                    opacity=alpha,
                    scene=ax['sceneN'],
                )

                fig.add_trace(trace)

    if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='normal'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)

    return fig, ax


from typing import Tuple, Union, Dict, Any, Literal

# General packages
import numpy as np
import plotly.graph_objects as go
import math

# Local packages
from comfit.tool import tool_colormap


superscripts = {'0': '⁰', '1': '¹', '2': '²', '3': '³', '4': '⁴',
                '5': '⁵', '6': '⁶', '7': '⁷', '8': '⁸', '9': '⁹',
                '-': '⁻'}

def format_tick_value(
    val: float
    ) -> str:
    """Format a numeric value with appropriate SI prefix and significant figures.
    This function formats numeric values using SI prefixes and adjusts decimal places
    based on the magnitude of the number. For values that don't align with standard
    SI prefixes, scientific notation is used as a fallback.

    Parameters
    ----------
    val : float
        The numeric value to format.

    Returns
    -------
    str
        A string representation of the value with appropriate SI prefix and formatting.
    """
    if val == 0:
        return '0'

    # prefixes = {exp: f'×10{"".join(superscripts[digit] for digit in str(exp))}' for exp in range(-18, 19, 3)}
    # prefix_letters = {
    #      -15: 'f', -12: 'p', -9: 'n', -6: 'µ', -3: 'm',
    #     0: '', 3: 'k', 6: 'M', 9: 'G', 12: 'T'
    # }
    # for key in prefix_letters.keys():
    #     prefixes[key] = prefix_letters[key]


    abs_val = abs(val)
    exp = int(np.floor(np.log10(abs_val) / 3) * 3)  # Round to nearest power of 1000
    
    prefixes = {}

    if exp in prefixes:
        scaled_val = val / 10**exp
        # Format based on the magnitude of scaled_val
        if abs(scaled_val) < 10:
            return f'{scaled_val:.2f}{prefixes[exp]}'  # 2.00m
        elif abs(scaled_val) < 100:
            return f'{scaled_val:.1f}{prefixes[exp]}'  # 20.0m
        else:
            return f'{scaled_val:.0f}{prefixes[exp]}'  # 200m
    else:
        return f'{val:.0e}'  # Fallback to scientific notation



def generate_numbers_between(
        cmin: float, 
        cmax: float
        ) -> Tuple[np.ndarray, float]:
    """Generate evenly spaced numbers between given minimum and maximum values.

    This function generates a sequence of numbers between cmin and cmax with
    appropriate step sizes based on the range magnitude. The step size is 
    adjusted to maintain a readable number of ticks.

    Parameters
    ----------
    cmin : float
        Minimum value of the range.
    cmax : float
        Maximum value of the range.

    Returns
    -------
    tuple[np.ndarray, float]
        A tuple containing:
        - np.ndarray: Array of evenly spaced numbers between cmin and cmax
        - float: The exponent of the range magnitude (delta_exp)

    Raises
    ------
    ValueError
        If cmin is greater than cmax.

    Notes
    -----
    The step size is determined based on the range magnitude:
    - If range <= 5 units: step = 10^delta_exp
    - If range <= 10 units: step = 2×10^delta_exp
    - If range <= 15 units: step = 3×10^delta_exp
    - If range <= 20 units: step = 4×10^delta_exp
    """
    if cmin > cmax:
        raise ValueError("cmin must be less than cmax")

    delta = (cmax - cmin)

    if delta == 0:
        return [0], 0
    
    delta_exp = np.floor(np.log10(max(abs(cmin), abs(cmax))))

    tick_min = np.ceil(cmin/10**delta_exp)*10**delta_exp
    tick_max = np.floor(cmax/10**delta_exp)*10**delta_exp

    number_of_steps = np.floor(delta/10**delta_exp)
    if number_of_steps <= 5:
        numbers_between = np.arange(tick_min, tick_max+10**delta_exp, 10**delta_exp)
    elif number_of_steps <= 10:
        numbers_between = np.arange(tick_min, tick_max+10**delta_exp, 2*10**delta_exp)
    elif number_of_steps <= 15:
        numbers_between = np.arange(tick_min, tick_max+10**delta_exp, 3*10**delta_exp)
    elif number_of_steps <= 20:
        numbers_between = np.arange(tick_min, tick_max+10**delta_exp, 4*10**delta_exp)

    return numbers_between, delta_exp
     
        

def tool_plotly_colorbar(
        ax : Dict[str, Any], 
        type : Literal['normal', 'angle'] = 'normal'
        ) -> Union[go.Scatter, go.Scatter3d]:
    """Add a colorbar to a plotly figure subplot with customizable type and placement.

    Parameters
    ----------
    ax : dict
        Dictionary containing subplot information with keys:
        - 'vmin': Minimum value for normal colorbar
        - 'vmax': Maximum value for normal colorbar
        - 'plot_dimension': Integer (2 or 3) indicating plot dimensions
        - 'colormap_object': Plotly colormap object
        - 'row': Row index of subplot
        - 'col': Column index of subplot
        - 'nrows': Total number of rows in figure
        - 'ncols': Total number of columns in figure

    type : str, optional
        Type of colorbar to create, by default 'normal'
        - 'normal': Linear scale with auto-formatted tick values
        - 'angle': Angular scale from -π to π with radian labels

    Returns
    -------
    Union[plotly.graph_objects.Scatter, plotly.graph_objects.Scatter3d]
        A scatter trace containing the colorbar configuration. The trace has no data points
        and is used solely for displaying the colorbar.

    Notes
    -----
    The colorbar placement is automatically calculated based on the subplot position.
    For normal type, tick values are auto-formatted with scientific notation.
    For angle type, tick values are displayed in π radians.
    """
 
    if type == 'normal':
        

        cmin = ax['vmin']
        cmax = ax['vmax']

        numbers_between = generate_numbers_between(cmin, cmax)

        tickvals, delta_exp = numbers_between
        
        ticktext = [round(tickval/10**delta_exp) for tickval in tickvals]

        title='×10' + ''.join([superscripts[digit] for digit in str(int(delta_exp))])
        
        # tickvals = np.linspace(cmin, cmax, 7)
        # ticktext = [format_tick_value(val) for val in tickvals]

    elif type == 'angle':
        title=None
        cmin = -np.pi
        cmax = np.pi
        tickvals = [-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi]
        ticktext = ['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
    
    if ax['plot_dimension'] == 2:
        trace = go.Scatter(
                x=[None], y=[None], mode='markers',
                showlegend=False,
                marker=dict(
                    colorscale=ax['colormap_object'],
                    cmin=cmin, 
                    cmax=cmax,
                    colorbar=dict(
                        tickvals=tickvals,
                        ticktext=ticktext
                    )
                ),
                hoverinfo='none'
                )
                
    elif ax['plot_dimension'] == 3:
        trace = go.Scatter3d(
                x=[None], y=[None], z=[None], mode='markers',
                showlegend=False,
                marker=dict(
                    colorscale=ax['colormap_object'],
                    cmin=cmin, 
                    cmax=cmax,
                    colorbar=dict(
                    tickvals=tickvals,
                    ticktext=ticktext
                    )
                ),
                hoverinfo='none'
                )
    
    # The idea here is to place the colorbar in the middle of the subplot
    # It is not perfect yet (Vidar, 13.02.25)
    # padding=0.0

    y_axis_placement = 1-1/ax['nrows']/2 -(ax['row']-1)/ax['nrows'] 
    #0.5 if nrows = 1, 0.33, 0.66 if nrows = 2, 0.25, 0.5, 0.75 if nrows = 3 etc. 
    y_axis_correction = 0 if ax['nrows'] == 1 else -(ax['row']-(1+ax['nrows'])/2)/(ax['nrows']-(1+ax['nrows'])/2) 
    #Goes from 1 (row=1) to -1 (row=nrows) (zero if nrows = 1)
    y_axis_constant = 0
                    
    trace.marker.colorbar.update(
        title=title,
            len=1/ax['nrows'],
            thickness=10,
            xanchor='left',
            x=ax['col']/ax['ncols']-0.1*(1-ax['col']/ax['ncols']),
            yanchor='middle',
            y= y_axis_placement + 0.03*y_axis_correction + y_axis_constant,
        )

    return trace


from typing import TYPE_CHECKING, List, Dict, Any
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem


def tool_extract_node_arrays(
        self : 'BaseSystem', 
        nodes : List[Dict]
        ) -> Dict[str, Any]:
    """Extracts node arrays from a dictionary of nodes.

    Parameters
    ----------
    self : 'BaseSystem'
        A BaseSystem (or derived) instance.
    nodes : List[Dict]
        List of node dictionaries containing position and other properties

    Returns
    -------
    Dict[str, Any]
        Dictionary containing arrays of node properties including:
        - coordinates (x,y,z)
        - velocity components if present
        - charge-separated coordinates if charges present
        - Burgers vector components if present
        - tangent vector components if present
    """

    node_arrays = {}

    # Check if there are nodes to be treated, if not return an empty dictionary
    if not nodes:
        return node_arrays
    
    node_arrays['velocity_given'] = 'velocity' in nodes[0].keys()

    x_coordinates = []
    y_coordinates = []
    z_coordinates = []

    x_coordinates_positive = []
    y_coordinates_positive = []
    z_coordinates_positive = []

    x_coordinates_negative = []
    y_coordinates_negative = []
    z_coordinates_negative = []

    node_arrays['charge_given'] = 'charge' in nodes[0].keys()
    node_arrays['velocity_given'] = 'velocity' in nodes[0].keys()
    node_arrays['Burgers_vector_given'] = 'Burgers_vector' in nodes[0].keys()
    node_arrays['tangent_vector_given'] = 'tangent_vector' in nodes[0].keys()

    for node in nodes:
        x_coordinates.append(node['position'][0])
        y_coordinates.append(node['position'][1] if cfi.dim > 1 else None)
        z_coordinates.append(node['position'][2] if cfi.dim > 2 else None)

    node_arrays['x_coordinates'] = x_coordinates
    node_arrays['y_coordinates'] = y_coordinates if cfi.dim > 1 else None
    node_arrays['z_coordinates'] = z_coordinates if cfi.dim > 2 else None

    if node_arrays['charge_given']:

        for node in nodes:

            if node['charge'] > 0:
                x_coordinates_positive.append(node['position'][0])
                y_coordinates_positive.append(node['position'][1] if cfi.dim > 1 else None)
                z_coordinates_positive.append(node['position'][2] if cfi.dim > 2 else None)

            else:
                x_coordinates_negative.append(node['position'][0])
                y_coordinates_negative.append(node['position'][1] if cfi.dim > 1 else None)
                z_coordinates_negative.append(node['position'][2] if cfi.dim > 2 else None)
            
        node_arrays['x_coordinates_positive'] = x_coordinates_positive
        node_arrays['y_coordinates_positive'] = y_coordinates_positive if cfi.dim > 1 else None
        node_arrays['z_coordinates_positive'] = z_coordinates_positive if cfi.dim > 2 else None

        node_arrays['x_coordinates_negative'] = x_coordinates_negative
        node_arrays['y_coordinates_negative'] = y_coordinates_negative if cfi.dim > 1 else None
        node_arrays['z_coordinates_negative'] = z_coordinates_negative if cfi.dim > 2 else None

    for property in ['velocity', 'tangent_vector', 'Burgers_vector']:
        node_arrays[property + '_given'] = property in nodes[0].keys()

        if node_arrays[property + '_given']:
            x_coordinates = []
            y_coordinates = []
            z_coordinates = []

            for node in nodes:
                x_coordinates.append(node[property][0])
                y_coordinates.append(node[property][1] if cfi.dim > 1 else None)
                z_coordinates.append(node[property][2] if cfi.dim > 2 else None)
            
            node_arrays[property + '_x_coordinates'] = x_coordinates
            node_arrays[property + '_y_coordinates'] = y_coordinates if cfi.dim > 1 else None
            node_arrays[property + '_z_coordinates'] = z_coordinates if cfi.dim > 2 else None

    return node_arrays
        
        



from typing import Optional

# General packages
import matplotlib.pyplot as plt

def tool_matplotlib_define_2D_plot_ax(
        fig : plt.Figure, 
        ax : Optional[plt.Axes] = None
        ) -> plt.Axes:
    """
    Defines the axes for a 2D plot in matplotlib.
    
    Parameters
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    ax : matplotlib.axes._subplots.AxesSubplot
        Axes object.
    
    Returns
    -------
    ax : matplotlib.axes._subplots.AxesSubplot
        Axes object.
    """

    if ax is None:
        ax = fig.add_subplot(111)
    
    return ax